# Copyright (C) 2025 Arcee AI
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import logging
from typing import List, Optional

import torch
import tqdm
import transformers

# explicitly import the config class so that we can catch errors upstream
# if the transformers version installed is too old
from transformers.models.qwen2_moe import Qwen2MoeConfig

from mergekitty.architecture import QWEN2_INFO
from mergekitty.moe.arch import MoEOutputArchitecture
from mergekitty.moe.common import copy_tensor_out, initialize_io, select_dtype
from mergekitty.moe.config import MoEMergeConfig
from mergekitty.options import MergeOptions


class QwenMoE(MoEOutputArchitecture):
    def name(self) -> str:
        return "Qwen MoE"

    def supports_config(
        self,
        config: MoEMergeConfig,
        explain: bool = False,
        trust_remote_code: bool = False,
    ) -> bool:
        if len(config.shared_experts or []) != 1:
            if explain:
                logging.warning("Qwen MoE merge requires exactly one shared expert")
            return False

        if (
            config.gate_mode != "random"
            and not config.shared_experts[0].positive_prompts
        ):
            if explain:
                logging.warning("Qwen MoE requires the shared expert to have prompts")
            return False

        model_types = []
        for model_ref in (
            [config.base_model]
            + [e.source_model for e in config.experts]
            + [e.source_model for e in (config.shared_experts or [])]
        ):
            model_cfg = model_ref.config(trust_remote_code=trust_remote_code)
            model_types.append(model_cfg.model_type)

        if len(set(model_types)) != 1:
            if explain:
                logging.warning(
                    "Qwen MoE requires all input models to have the same architecture"
                )
            return False
        if model_types[0] not in ("llama", "mistral", "qwen2"):
            if explain:
                logging.warning(
                    "Qwen MoE requires all input models to be Qwen2, Llama or Mistral models"
                )
            return False
        return True

    def _generate_config(
        self,
        base_config: transformers.PretrainedConfig,
        num_experts: int,
        experts_per_token: Optional[int] = None,
    ) -> Qwen2MoeConfig:
        out_cfg = Qwen2MoeConfig(**base_config.to_dict())
        out_cfg.architectures = ["Qwen2MoeForCausalLM"]
        out_cfg.num_experts = num_experts
        out_cfg.num_experts_per_tok = experts_per_token or 2
        out_cfg.decoder_sparse_step = 1
        out_cfg.norm_topk_prob = True
        out_cfg.sliding_window = None
        out_cfg.use_sliding_window = False
        out_cfg.shared_expert_intermediate_size = out_cfg.intermediate_size
        out_cfg.moe_intermediate_size = out_cfg.intermediate_size

        if (out_cfg.num_experts & (out_cfg.num_experts - 1)) != 0:
            logging.warning(
                f"Your model has {out_cfg.num_experts} experts, which is "
                "not a power of two. The model will not be usable in llama.cpp."
            )
        return out_cfg

    def write_model(
        self,
        out_path: str,
        config: MoEMergeConfig,
        merge_options: MergeOptions,
        router_weights: List[torch.Tensor],
        shared_router_weights: Optional[List[torch.Tensor]] = None,
    ):
        base_model = config.base_model
        base_cfg = base_model.config(trust_remote_code=merge_options.trust_remote_code)

        out_dtype = select_dtype(config, base_cfg)
        out_cfg = self._generate_config(
            base_cfg,
            len(config.experts),
            config.experts_per_token,
        )
        if out_dtype is not None:
            out_cfg.torch_dtype = out_dtype
        out_cfg.save_pretrained(out_path)

        shared_def = config.shared_experts[0]

        loaders, base_loader, writer = initialize_io(config, out_path, merge_options)
        shared_loader = loaders.get(shared_def.source_model) if shared_def else None
        for weight_info in tqdm.tqdm(
            QWEN2_INFO.all_weights(base_cfg),
            desc="Weights",
        ):
            tensor_name = weight_info.name
            if ".mlp." in tensor_name:
                for expert_idx, expert in enumerate(config.experts):
                    expert_name = tensor_name.replace(
                        ".mlp.", f".mlp.experts.{expert_idx}."
                    )
                    expert_loader = loaders.get(expert.source_model)
                    copy_tensor_out(
                        weight_info,
                        expert_loader,
                        writer,
                        expert=expert,
                        is_residual="down_proj" in tensor_name,
                        output_name=expert_name,
                        out_dtype=out_dtype,
                        clone=merge_options.clone_tensors,
                    )

                copy_tensor_out(
                    weight_info,
                    shared_loader,
                    writer,
                    expert=shared_def,
                    is_residual="down_proj" in tensor_name,
                    output_name=tensor_name.replace(".mlp.", ".mlp.shared_expert."),
                    out_dtype=out_dtype,
                    clone=merge_options.clone_tensors,
                )
            else:
                try:
                    tensor = base_loader.get_tensor(
                        tensor_name, aliases=weight_info.aliases
                    )
                except KeyError:
                    if tensor_name.endswith("_proj.bias"):
                        # qwen 2 moe wants attention bias, give it zeros
                        head_dim = out_cfg.hidden_size // out_cfg.num_attention_heads
                        num_heads = (
                            out_cfg.num_key_value_heads
                            if (
                                tensor_name.endswith("k_proj.bias")
                                or tensor_name.endswith("v_proj.bias")
                            )
                            else out_cfg.num_attention_heads
                        )
                        tensor = torch.zeros(num_heads * head_dim, dtype=out_dtype)
                    elif weight_info.optional:
                        continue
                    else:
                        raise

                writer.save_tensor(
                    tensor_name,
                    tensor.to(dtype=out_dtype),
                    clone=merge_options.clone_tensors,
                )

        for layer_idx, weight in enumerate(
            tqdm.tqdm(router_weights, desc="Router weights")
        ):
            writer.save_tensor(
                f"model.layers.{layer_idx}.mlp.gate.weight",
                weight.to(dtype=out_dtype).contiguous(),
                clone=merge_options.clone_tensors,
            )
            writer.save_tensor(
                f"model.layers.{layer_idx}.mlp.shared_expert_gate.weight",
                shared_router_weights[layer_idx].to(dtype=out_dtype).contiguous(),
                clone=merge_options.clone_tensors,
            )

        writer.finalize()



# Copyright (C) 2025 Arcee AI
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import logging
from typing import List, Optional

import torch
import tqdm
import transformers

from mergekitty.architecture import MISTRAL_INFO, WeightInfo
from mergekitty.moe.arch import MoEOutputArchitecture
from mergekitty.moe.common import copy_tensor_out, initialize_io, select_dtype
from mergekitty.moe.config import MoEMergeConfig
from mergekitty.options import MergeOptions


class MixtralMoE(MoEOutputArchitecture):
    def name(self) -> str:
        return "Mixtral"

    def supports_config(
        self,
        config: MoEMergeConfig,
        explain: bool = False,
        trust_remote_code: bool = False,
    ) -> bool:
        if config.shared_experts:
            if explain:
                logging.warning("Mixtral does not support shared experts")
            return False

        model_types = []
        for model_ref in [config.base_model] + [e.source_model for e in config.experts]:
            model_cfg = model_ref.config(trust_remote_code=trust_remote_code)
            model_types.append(model_cfg.model_type)

        if len(set(model_types)) != 1:
            if explain:
                logging.warning(
                    "Mixtral requires all input models to have the same architecture"
                )
            return False
        if model_types[0] not in ("llama", "mistral"):
            if explain:
                logging.warning(
                    "Mixtral requires all input models to be Llama or Mistral models"
                )
            return False
        return True

    def _generate_config(
        self,
        base_config: transformers.PretrainedConfig,
        num_experts: int,
        shared_experts: Optional[int] = None,
        experts_per_token: Optional[int] = None,
    ) -> transformers.PretrainedConfig:
        if shared_experts:
            raise NotImplementedError("Shared experts not supported for Mixtral output")

        if not isinstance(base_config, transformers.MistralConfig):
            base_cfg_mistral = transformers.MistralConfig(**base_config.to_dict())
            base_cfg_mistral.sliding_window = None
            base_cfg_mistral.max_position_embeddings = (
                base_config.max_position_embeddings
            )
            base_config = base_cfg_mistral

        out_cfg = transformers.MixtralConfig(**base_config.to_dict())
        out_cfg.architectures = ["MixtralForCausalLM"]
        out_cfg.num_local_experts = num_experts
        out_cfg.num_experts_per_tok = experts_per_token or 2
        out_cfg.sliding_window = None

        if (out_cfg.num_local_experts & (out_cfg.num_local_experts - 1)) != 0:
            logging.warning(
                f"Your model has {out_cfg.num_local_experts} experts, which is "
                "not a power of two. The model will not be usable in llama.cpp."
            )
        return out_cfg

    def _remap_weight_name(self, weight: WeightInfo) -> str:
        if ".mlp." not in weight.name:
            # Everything but MLP is identical to base Mistral
            return weight.name

        res = weight.name
        for needle, replacement in [
            (".mlp.gate_proj", ".block_sparse_moe.experts.{expert_idx}.w1"),
            (".mlp.down_proj", ".block_sparse_moe.experts.{expert_idx}.w2"),
            (".mlp.up_proj", ".block_sparse_moe.experts.{expert_idx}.w3"),
        ]:
            res = res.replace(needle, replacement)
        return res

    def _router_weight_name(self, layer_idx: int) -> str:
        return f"model.layers.{layer_idx}.block_sparse_moe.gate.weight"

    def write_model(
        self,
        out_path: str,
        config: MoEMergeConfig,
        merge_options: MergeOptions,
        router_weights: List[torch.Tensor],
        shared_router_weights: Optional[List[torch.Tensor]] = None,
    ):
        base_model = config.base_model
        base_cfg = base_model.config(trust_remote_code=merge_options.trust_remote_code)

        assert len(router_weights) == base_cfg.num_hidden_layers, (
            f"Expected {base_cfg.num_hidden_layers} router weights, "
            f"got {len(router_weights)}"
        )

        out_dtype = select_dtype(config, base_cfg)
        out_cfg = self._generate_config(
            base_cfg,
            len(config.experts),
            len(config.shared_experts or []),
            config.experts_per_token,
        )
        out_cfg.torch_dtype = out_dtype
        out_cfg.save_pretrained(out_path)

        loaders, base_loader, writer = initialize_io(config, out_path, merge_options)
        for weight_info in tqdm.tqdm(
            MISTRAL_INFO.all_weights(base_cfg),
            desc="Weights",
        ):
            tensor_name = self._remap_weight_name(weight_info)
            if "{expert_idx}" in tensor_name:
                for expert_index, expert in enumerate(config.experts):
                    expert_name = tensor_name.replace("{expert_idx}", str(expert_index))
                    expert_loader = loaders.get(expert.source_model)
                    copy_tensor_out(
                        weight_info,
                        expert_loader,
                        writer,
                        expert=expert,
                        out_dtype=out_dtype,
                        output_name=expert_name,
                        clone=merge_options.clone_tensors,
                        is_residual="down_proj" in tensor_name,
                    )
            else:
                copy_tensor_out(
                    weight_info,
                    base_loader,
                    writer,
                    out_dtype=out_dtype,
                    clone=merge_options.clone_tensors,
                )

        for layer_idx, weight in enumerate(
            tqdm.tqdm(router_weights, desc="Router weights")
        ):
            writer.save_tensor(
                self._router_weight_name(layer_idx),
                weight.to(dtype=out_dtype).contiguous(),
                clone=merge_options.clone_tensors,
            )

        writer.finalize()



# Copyright (C) 2025-2026 Allura-org

try:
    import gradio as gr
    from gradio_log import Log
    from gradio_huggingfacehub_search import HuggingfaceHubSearch
except ImportError:
    raise ImportError(
        "gradio is not installed. Please install mergekitty with `pip install mergekitty[gradio]`."
    )
from mergekitty.merge import MergeConfiguration, MergeOptions, run_merge
from mergekitty.merge_methods import REGISTERED_MERGE_METHODS
import torch
import yaml
import huggingface_hub as hf
import click
import tempfile
from contextlib import redirect_stdout, redirect_stderr

from mergekitty.merge_methods.base import MergeMethod

PROD = True


def do_merge(
    token: gr.OAuthToken, yaml_config: str, log_file: str, output_model_path: str
):
    config = MergeConfiguration.model_validate(yaml.safe_load(yaml_config))
    options = MergeOptions(
        allow_crimes=False,
        allow_remote_code=True,
        lazy_unpickle=True,
        cuda=torch.cuda.is_available(),
    )
    with open(log_file, "w") as f:
        with redirect_stdout(f), redirect_stderr(f):
            temp_dir = tempfile.mkdtemp()
            run_merge(config, temp_dir, options)
            hf.create_repo(output_model_path, repo_type="model", private=True)
            hf.upload_folder(
                repo_id=output_model_path,
                folder_path=temp_dir,
                repo_type="model",
                token=token.token,
            )
            gr.Info(
                f"Model uploaded to <a href='https://huggingface.co/{output_model_path}'>{output_model_path}</a> successfully!"
            )
    with open(log_file, "w") as f:
        f.write("Merge complete!")
    return


def main(share: bool):
    tmp = tempfile.NamedTemporaryFile(delete=False)
    with gr.Blocks() as gui:
        state = gr.State(value=tmp.name)
        gr.Markdown("# Mergekitty GUI")
        with gr.Tab("YAML config"):
            with gr.Row():
                with gr.Column():
                    output_model_path = gr.Textbox(
                        label="Output model path (Huggingface identifier, ie `estrogen/test-model` -- this should be a user or org you have write access to)",
                        value="estrogen/test-model",
                    )
                    yaml_config = gr.Code(
                        language="yaml",
                    )
                    btn_yaml = gr.Button("Merge!")

                with gr.Column():
                    Log(log_file=tmp.name, dark=True, label="Logs", container=True)
        with gr.Tab("GUI config"):
            if PROD:
                with gr.Row():
                    gr.Markdown("TODO!")
            else:
                with gr.Column():
                    with gr.Row():
                        with gr.Column():
                            output_model_path = gr.Textbox(
                                label="Output model path (Huggingface identifier, ie `estrogen/test-model` -- this should be a user or org you have write access to)",
                                value="estrogen/test-model",
                            )
                            HuggingfaceHubSearch(label="Base model", interactive=True)
                            merge_method = gr.Dropdown(
                                label="Merge method",
                                choices=list(REGISTERED_MERGE_METHODS.keys()),
                                value=list(REGISTERED_MERGE_METHODS.keys())[0],
                                interactive=True,
                            )
                        with gr.Column():
                            models = []

                            def add_model(model: str):
                                models.append({"model": model, "parameters": {}})
                                return ""  # blank out the search box

                            with gr.Row():
                                gui_search = HuggingfaceHubSearch(
                                    label="Search for a model", interactive=True
                                )
                                btn_gui_search = gr.Button("Add model")
                                btn_gui_search.click(
                                    add_model, inputs=[gui_search], outputs=[gui_search]
                                )

                            @gr.render(
                                inputs=[gui_search, merge_method],
                                triggers=[btn_gui_search.click, merge_method.change],
                            )
                            def render_models(model: str, merge_method: str):
                                merge_method: MergeMethod = REGISTERED_MERGE_METHODS[
                                    merge_method
                                ]
                                tensor_params = merge_method.tensor_parameters()
                                for model in models:

                                    def remove_model():
                                        models.remove(model)

                                    with gr.Row(variant="panel"):
                                        with gr.Column():
                                            HuggingfaceHubSearch(
                                                label="Model", value=model["model"]
                                            )
                                            with gr.Accordion("Parameters", open=False):
                                                for param in tensor_params:

                                                    def update_params(param_value: str):
                                                        model["parameters"][
                                                            param.name
                                                        ] = param_value

                                                    box = gr.Textbox(
                                                        label=param.name,
                                                        value=param.default_value,
                                                        interactive=True,
                                                    )
                                                    box.change(
                                                        update_params, inputs=[box]
                                                    )
                                        btn = gr.Button("Remove model")
                                        btn.click(remove_model)

                    with gr.Column():
                        btn_gui = gr.Button("Merge!")
                        Log(log_file=tmp.name, dark=True, label="Logs", container=True)
        gr.LoginButton()

        btn_yaml.click(do_merge, inputs=[yaml_config, state, output_model_path])
        if not PROD:
            btn_gui.click(lambda: print(models))

    tmp.close()

    gui.queue(default_concurrency_limit=1).launch(share=share)


@click.command("mergekitty-gui")
@click.option(
    "--share", is_flag=True, required=False, default=False, help="Share the GUI"
)
def cli_main(share: bool):
    main(share)


if __name__ == "__main__":
    cli_main()



# Partly Copyright (C) 2025 Arcee AI
# Partly Copyright (C) 2025-2026 Allura-org
import threading
from types import SimpleNamespace
from typing import Any, Dict, Optional

import networkx
import pytest

from mergekitty.common import ImmutableMap, ModelReference
from mergekitty.executor import ParallelismExecutor, SingleThreadedExecutor
from mergekitty.io.tasks import LoadTensor, LoaderCache
from mergekitty.task import Task

EXECUTION_COUNTS: Dict[Task, int] = {}
BLOCKING_TASKS_STARTED: set[str] = set()
BLOCKING_TASKS_LOCK = threading.Lock()
BLOCKING_TASKS_READY = threading.Event()
BLOCKING_TASKS_RELEASE = threading.Event()


class DummyTask(Task):
    result: Any
    dependencies: ImmutableMap[str, Task]
    name: str = "DummyTask"
    grouplabel: Optional[str] = None
    execution_count: int = 0

    def arguments(self):
        return self.dependencies

    def group_label(self) -> Optional[str]:
        return self.grouplabel

    def execute(self, **kwargs):
        EXECUTION_COUNTS[self] = EXECUTION_COUNTS.get(self, 0) + 1
        return self.result


def create_mock_task(name, result=None, dependencies=None, group_label=None):
    if dependencies is None:
        dependencies = {}
    return DummyTask(
        result=result,
        dependencies=ImmutableMap(data=dependencies),
        name=name,
        grouplabel=group_label,
    )


@pytest.fixture(
    params=[SingleThreadedExecutor, ParallelismExecutor],
    ids=["single", "parallel"],
)
def executor_cls(request):
    return request.param


# Test cases for the Task implementation
class TestTaskClass:
    def test_task_execute(self):
        # Testing the execute method
        task = create_mock_task("task1", result=42)
        assert task.execute() == 42, "Task execution did not return expected result"

    def test_task_priority(self):
        task = create_mock_task("task1")
        assert task.priority() == 0, "Default priority should be 0"

    def test_task_group_label(self):
        task = create_mock_task("task1")
        assert task.group_label() is None, "Default group label should be None"


# Test cases for the Executor implementation
class TestExecutorClass:
    def test_executor_initialization(self, executor_cls):
        # Testing initialization with single task
        task = create_mock_task("task1")
        executor = executor_cls([task])
        assert executor.targets == [task], (
            "Executor did not initialize with correct targets"
        )

    def test_executor_empty_list(self, executor_cls):
        list(executor_cls([]).run())

    def test_executor_scheduling(self, executor_cls):
        # Testing scheduling with dependencies
        task1 = create_mock_task("task1", result=1)
        task2 = create_mock_task("task2", result=2, dependencies={"task1": task1})
        executor = executor_cls([task2])
        assert len(executor._make_schedule([task2])) == 2, (
            "Schedule should include two tasks"
        )

    def test_executor_dependency_building(self, executor_cls):
        # Testing dependency building
        task1 = create_mock_task("task1")
        task2 = create_mock_task("task2", dependencies={"task1": task1})
        executor = executor_cls([task2])
        dependencies = executor._build_dependencies([task2])
        assert task1 in dependencies[task2], "Task1 should be a dependency of Task2"

    def test_executor_run(self, executor_cls):
        # Testing execution through the run method
        task1 = create_mock_task("task1", result=10)
        task2 = create_mock_task("task2", result=20, dependencies={"task1": task1})
        executor = executor_cls([task2])
        results = list(executor.run())
        assert len(results) == 1 and results[0][1] == 20, (
            "Executor run did not yield correct results"
        )

    def test_executor_execute(self, executor_cls):
        # Testing execute method for side effects
        task1 = create_mock_task("task1", result=10)
        executor = executor_cls([task1])
        # No assert needed; we're ensuring no exceptions are raised and method completes
        executor.execute()

    def test_dependency_ordering(self, executor_cls):
        # Testing the order of task execution respects dependencies
        task1 = create_mock_task("task1", result=1)
        task2 = create_mock_task("task2", result=2, dependencies={"task1": task1})
        task3 = create_mock_task("task3", result=3, dependencies={"task2": task2})
        executor = executor_cls([task3])

        schedule = executor._make_schedule([task3])
        assert schedule.index(task1) < schedule.index(task2), (
            "Task1 should be scheduled before Task2"
        )
        assert schedule.index(task2) < schedule.index(task3), (
            "Task2 should be scheduled before Task3"
        )


class TestExecutorGroupLabel:
    def test_group_label_scheduling(self, executor_cls):
        # Create tasks with group labels and dependencies
        task1 = create_mock_task("task1", group_label="group1")
        task2 = create_mock_task(
            "task2", dependencies={"task1": task1}, group_label="group1"
        )
        task3 = create_mock_task("task3", group_label="group2")
        task4 = create_mock_task(
            "task4", dependencies={"task2": task2, "task3": task3}, group_label="group1"
        )

        # Initialize Executor with the tasks
        executor = executor_cls([task4])

        # Get the scheduled tasks
        schedule = executor._make_schedule([task4])

        # Check if tasks with the same group label are scheduled consecutively when possible
        group_labels_in_order = [
            task.group_label() for task in schedule if task.group_label()
        ]
        assert group_labels_in_order == [
            "group1",
            "group1",
            "group2",
            "group1",
        ], "Tasks with same group label are not scheduled consecutively"

    def test_group_label_with_dependencies(self, executor_cls):
        # Creating tasks with dependencies and group labels
        task1 = create_mock_task("task1", result=1, group_label="group1")
        task2 = create_mock_task(
            "task2", result=2, dependencies={"task1": task1}, group_label="group2"
        )
        task3 = create_mock_task(
            "task3", result=3, dependencies={"task2": task2}, group_label="group1"
        )

        executor = executor_cls([task3])
        schedule = executor._make_schedule([task3])
        scheduled_labels = [
            task.group_label() for task in schedule if task.group_label()
        ]

        # Check if task3 is scheduled after task1 and task2 due to dependency, even though it has the same group label as task1
        group1_indices = [
            i for i, label in enumerate(scheduled_labels) if label == "group1"
        ]
        group2_index = scheduled_labels.index("group2")

        assert group1_indices[-1] > group2_index, (
            "Task with the same group label but later dependency was not scheduled after different group label"
        )

    def test_load_tensor_scheduling_groups_by_shard(self, executor_cls, monkeypatch):
        model = ModelReference.parse("fake-model")
        tensor_paths = {
            "model.layers.0.self_attn.q_proj.weight": "model-00001-of-00004.safetensors",
            "model.layers.1.self_attn.q_proj.weight": "model-00001-of-00004.safetensors",
            "model.layers.2.self_attn.q_proj.weight": "model-00001-of-00004.safetensors",
            "model.layers.10.self_attn.q_proj.weight": "model-00003-of-00004.safetensors",
            "model.layers.11.self_attn.q_proj.weight": "model-00003-of-00004.safetensors",
            "model.layers.12.self_attn.q_proj.weight": "model-00003-of-00004.safetensors",
        }
        fake_loader = SimpleNamespace(
            index=SimpleNamespace(tensor_paths=tensor_paths),
        )
        monkeypatch.setattr(LoaderCache, "get", lambda self, _model: fake_loader)

        tasks = [LoadTensor(model=model, tensor=name) for name in tensor_paths]
        schedule = executor_cls(tasks)._make_schedule(tasks)

        shard_switches = 0
        previous_shard = None
        for task in schedule:
            shard = tensor_paths[task.tensor]
            if previous_shard is not None and shard != previous_shard:
                shard_switches += 1
            previous_shard = shard

        assert shard_switches <= 1, (
            "LoadTensor scheduling should preserve shard locality"
        )


class TestExecutorSingleExecution:
    def test_single_execution_per_task(self, executor_cls):
        EXECUTION_COUNTS.clear()

        shared_task = create_mock_task("shared_task", result=100)
        task1 = create_mock_task("task1", dependencies={"shared": shared_task})
        task2 = create_mock_task("task2", dependencies={"shared": shared_task})
        task3 = create_mock_task("task3", dependencies={"task1": task1, "task2": task2})

        executor_cls([task3]).execute()

        assert shared_task in EXECUTION_COUNTS, "Dependency not executed"
        assert EXECUTION_COUNTS[shared_task] == 1, (
            "Shared dependency should be executed exactly once"
        )


class CircularTask(Task):
    def arguments(self) -> Dict[str, Task]:
        return {"its_a_me": self}

    def execute(self, **_kwargs) -> Any:
        assert False, "Task with circular dependency executed"


class BlockingTask(Task):
    name: str

    def arguments(self) -> Dict[str, Task]:
        return {}

    def execute(self, **_kwargs) -> str:
        with BLOCKING_TASKS_LOCK:
            BLOCKING_TASKS_STARTED.add(self.name)
            if len(BLOCKING_TASKS_STARTED) >= 2:
                BLOCKING_TASKS_READY.set()

        assert BLOCKING_TASKS_RELEASE.wait(timeout=5), (
            "Timed out waiting for blocking tasks to be released"
        )
        return self.name


class TestExecutorCircularDependency:
    def test_circular_dependency(self, executor_cls):
        with pytest.raises(networkx.NetworkXUnfeasible):
            executor_cls([CircularTask()]).execute()


class TestParallelismExecutor:
    def test_executes_ready_tasks_in_parallel(self):
        with BLOCKING_TASKS_LOCK:
            BLOCKING_TASKS_STARTED.clear()
        BLOCKING_TASKS_READY.clear()
        BLOCKING_TASKS_RELEASE.clear()

        left = BlockingTask(name="left")
        right = BlockingTask(name="right")
        target = create_mock_task(
            "target",
            dependencies={"left": left, "right": right},
        )

        executor = ParallelismExecutor([target], max_workers=2)
        errors = []

        def _run_executor():
            try:
                executor.execute()
            except Exception as exc:  # pragma: no cover - assertion handles this
                errors.append(exc)
                BLOCKING_TASKS_RELEASE.set()

        thread = threading.Thread(target=_run_executor, daemon=True)
        thread.start()

        assert BLOCKING_TASKS_READY.wait(timeout=5), (
            "Parallel executor did not start both ready tasks concurrently"
        )
        with BLOCKING_TASKS_LOCK:
            assert BLOCKING_TASKS_STARTED == {"left", "right"}

        BLOCKING_TASKS_RELEASE.set()
        thread.join(timeout=5)

        assert not thread.is_alive(), "Parallel executor did not finish"
        assert not errors


if __name__ == "__main__":
    pytest.main()



#!/usr/bin/env python3
"""
Merges multiple models and their dependencies into a single model
using multiple merge yaml documents in a single yaml file as the input
"""

import logging
import os
import sys
from pathlib import Path

import click
import yaml

from mergekitty.config import MergeConfiguration
from mergekitty.merge import MergeOptions, run_merge
from mergekitty.options import add_merge_options

merges = {}


def has_circular_dependency(nodes):
    """
    Detects circular in merges dependencies using DFS
    Returns the node where the cycle is detected
    """

    def dfs(node, visited, stack):
        """
        Returns True if a cycle is detected
        """
        visited[node] = True
        stack[node] = True

        for dependency in nodes[node]["deps"]:
            if not visited[dependency]:
                if dfs(dependency, visited, stack):
                    return True
            elif stack[dependency]:
                return True

        stack[node] = False
        return False

    visited = {key: False for key in nodes}
    stack = {key: False for key in nodes}

    for node in nodes:
        if not visited[node]:
            if dfs(node, visited, stack):
                return node

    return None


def merge(m: str, merge_options: MergeOptions, force: bool, out_path: Path):
    """
    Merges a model and its dependencies

    Params:
        m: name of the model to merge
        merge_options: MergeOptions
        force: overwrite existing merge results
        out_path: output path
    """
    # check if output_path exists
    if os.path.exists(out_path / m):
        if not force:
            logging.info("Skipping %s as it already exists", m)
            del merges[m]
            return
        logging.info("Overwriting %s as --force was specified", m)

    if len(merges[m]["deps"]) != 0:
        for dep in merges[m]["deps"]:
            if dep in merges:
                merge(dep, merge_options, force, out_path)

    logging.info("Merging model %s", m)
    merge_config: MergeConfiguration = MergeConfiguration.model_validate(merges[m])
    run_merge(
        merge_config,
        str(out_path / merges[m]["name"]),
        options=merge_options,
    )
    del merges[m]


def add_model_deps(model: str, name: str, out_path: Path):
    """
    Adds a model to `name`s dependencies if it is not already there and is a merge
    """
    model_lora = model.split("+")
    # name must not have a slash to avoid path traversal
    # therefore, we can use it to check if its a merge from the config
    if "/" not in model_lora[0]:
        # avoid duplicate deps
        if model_lora[0] not in merges[name]["deps"]:
            merges[name]["deps"].append(model_lora[0])
        model = str(out_path / model_lora[0])
        if len(model_lora) == 2:
            model += "+" + model_lora[1]

    return model


@click.command("mergekitty-mega")
@click.argument("config_file")
@click.argument("out_path")
@click.option(
    "--verbose", "-v", type=bool, default=False, is_flag=True, help="Verbose logging"
)
@click.option(
    "--force",
    "-f",
    type=bool,
    default=False,
    is_flag=True,
    help="Overwrite existing merge results instead of skipping them",
)
@click.option(
    "--require-nameless",
    "-R",
    type=bool,
    default=False,
    is_flag=True,
    help="Enforces exactly one unnamed merge in the YAML, which will inherit the input file's name.",
)
@add_merge_options
def main(
    merge_options: MergeOptions,
    config_file: str,
    out_path: str,
    force: bool,
    verbose: bool,
    require_nameless: bool,
):
    """
    Main entrypoint for mergekitty-mega command see module docstring for more info
    Params are supplied by click decorators
    """
    logging.basicConfig(level=logging.INFO if verbose else logging.WARNING)

    out_path = Path(out_path)
    final_found = False

    with open(config_file, "r", encoding="utf-8") as f:
        data = yaml.load_all(f, Loader=yaml.FullLoader)

        for d in data:
            if "name" not in d:
                if final_found:
                    logging.error("Only one merge must not have a name")
                    sys.exit(1)
                # this sets the name of the final merge to the config file name without the extension
                d["name"] = os.path.basename(config_file).rsplit(".", maxsplit=1)[0]
                final_found = True

            if "/" in d["name"]:
                logging.error("name must not contain a slash")
                sys.exit(1)

            merges[d["name"]] = d
            merges[d["name"]]["deps"] = []
            if "base_model" in d:
                d["base_model"] = add_model_deps(d["base_model"], d["name"], out_path)
            if "slices" in d:
                for slc in d["slices"]:
                    for src in slc["sources"]:
                        src["model"] = add_model_deps(src["model"], d["name"], out_path)
            if "models" in d:
                for mdl in d["models"]:
                    mdl["model"] = add_model_deps(mdl["model"], d["name"], out_path)

    if require_nameless and not final_found:
        logging.error("No final merge found")
        sys.exit(1)

    logging.info("Merging: %s", ", ".join(merges))

    if (dep := has_circular_dependency(merges)) is not None:
        logging.error("Circular dependency detected: %s", dep)
        sys.exit(1)

    while len(merges) != 0:
        m = list(merges.keys())[0]
        merge(m, merge_options, force, out_path)


if __name__ == "__main__":
    main()



"""Custom task implementations for mergekitty."""

import math
from typing import Dict

import torch

from mergekitty.task import Task


class MergeLoraTask(Task[torch.Tensor]):
    """Merge LoRA tensors into a base weight."""

    base_tensor: Task
    adapter_tensor_a: Task
    adapter_tensor_b: Task
    lora_alpha: float
    use_rslora: bool = False
    fan_in_fan_out: bool = False
    is_embedding: bool = False

    def arguments(self) -> Dict[str, Task]:
        return {
            "base_tensor": self.base_tensor,
            "adapter_tensor_a": self.adapter_tensor_a,
            "adapter_tensor_b": self.adapter_tensor_b,
        }

    def execute(self, **kwargs) -> torch.Tensor:
        base_tensor = kwargs["base_tensor"]
        adapter_tensor_a = kwargs["adapter_tensor_a"]
        adapter_tensor_b = kwargs["adapter_tensor_b"]

        rank = adapter_tensor_a.shape[0]
        if rank <= 0:
            raise RuntimeError("LoRA rank must be positive")

        scale = self.lora_alpha / (math.sqrt(rank) if self.use_rslora else rank)

        old_dtype = base_tensor.dtype
        base_fp32 = base_tensor.to(torch.float32)
        tensor_a_fp32 = adapter_tensor_a.to(torch.float32)
        tensor_b_fp32 = adapter_tensor_b.to(torch.float32)

        delta = tensor_b_fp32 @ tensor_a_fp32
        if self.is_embedding or self.fan_in_fan_out:
            delta = delta.transpose(0, 1)

        if delta.shape != base_fp32.shape:
            raise RuntimeError(
                f"LoRA delta shape {delta.shape} does not match base tensor shape "
                f"{base_fp32.shape}"
            )

        return (base_fp32 + (delta * scale)).to(old_dtype)

    def uses_accelerator(self) -> bool:
        return True



# Partly Copyright (C) 2025 Arcee AI
# Partly Copyright (C) 2025-2026 Allura-org
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

from abc import ABC, abstractmethod
from typing import Dict, Optional, Sequence

import safetensors
import torch

from mergekitty.io.lazy_unpickle import (
    DeferredLoad,
    TorchArchiveReader,
    torch_lazy_load,
)


class TensorLoader(ABC):
    """Base class for (potentially lazy) tensor loaders."""

    @abstractmethod
    def get_tensor(self, key: str) -> torch.Tensor: ...

    @abstractmethod
    def keys(self) -> Sequence[str]: ...

    @classmethod
    def get(
        cls,
        shard_path: str,
        use_lazy_unpickle: bool = False,
        device: Optional[str] = None,
    ) -> "TensorLoader":
        if shard_path.lower().endswith(".safetensors"):
            # not a subclass of TensorLoader, but exposes same api
            return safetensors.safe_open(
                shard_path, framework="pt", device=device or "cpu"
            )
        elif use_lazy_unpickle:
            return LazyPickleLoader(shard_path, device=device)
        return DumbPytorchLoader(shard_path, device=device)


class LazyPickleLoader(TensorLoader):
    """Loader for pytorch files using a custom unpickler and vigorous monkeypatching."""

    zip_reader: TorchArchiveReader
    index: Dict[str, DeferredLoad | torch.Tensor]
    device: Optional[str] = None

    def __init__(self, path: str, device: Optional[str] = None):
        self.zip_reader = TorchArchiveReader(path)
        self.device = device
        with torch_lazy_load():
            # Our custom unpickler relies on the legacy pickle path rather than
            # torch.load's eager weights-only loader.
            self.index = torch.load(path, weights_only=False)

    def get_tensor(self, key: str) -> torch.Tensor:
        if key not in self.index:
            raise KeyError(key)

        value = self.index[key]
        if isinstance(value, DeferredLoad):
            return value.execute(self.zip_reader, map_location=self.device)
        if isinstance(value, torch.Tensor):
            if self.device is not None and value.device != torch.device(self.device):
                return value.to(self.device)
            return value

        raise TypeError(
            f"Unexpected lazy-unpickle entry type for {key}: {type(value).__name__}"
        )

    def keys(self) -> Sequence[str]:
        return self.index.keys()


class DumbPytorchLoader(TensorLoader):
    """Naive `torch.load` shard loading."""

    tensors: Dict[str, torch.Tensor]

    def __init__(self, path: str, device: Optional[str] = None):
        self.tensors = torch.load(path, map_location=device, weights_only=True)

    def get_tensor(self, key: str) -> torch.Tensor:
        return self.tensors[key]

    def keys(self) -> Sequence[str]:
        return self.tensors.keys()



import pytest

from mergekitty.common import ModelPath, ModelReference


class TestModelReference:
    def test_parse_simple(self):
        text = "hf_user/model"
        mr = ModelReference.parse(text)
        assert mr.model == ModelPath(path="hf_user/model", revision=None)
        assert mr.lora is None
        assert str(mr) == text

    def test_parse_lora(self):
        text = "hf_user/model+hf_user/lora"
        mr = ModelReference.parse(text)
        assert mr.model == ModelPath(path="hf_user/model", revision=None)
        assert mr.lora == ModelPath(path="hf_user/lora", revision=None)
        assert str(mr) == text

    def test_parse_revision(self):
        text = "hf_user/model@v0.0.1"
        mr = ModelReference.parse(text)
        assert mr.model == ModelPath(path="hf_user/model", revision="v0.0.1")
        assert mr.lora is None
        assert str(mr) == text

    def test_parse_lora_plus_revision(self):
        text = "hf_user/model@v0.0.1+hf_user/lora@main"
        mr = ModelReference.parse(text)
        assert mr.model == ModelPath(path="hf_user/model", revision="v0.0.1")
        assert mr.lora == ModelPath(path="hf_user/lora", revision="main")
        assert str(mr) == text

    def test_parse_bad(self):
        with pytest.raises(RuntimeError):
            ModelReference.parse("@@@@@")

        with pytest.raises(RuntimeError):
            ModelReference.parse("a+b+c")

        with pytest.raises(RuntimeError):
            ModelReference.parse("a+b+c@d+e@f@g")



# Partly Copyright (C) 2025 Arcee AI
# Partly Copyright (C) 2025-2026 Allura-org

import logging
import os
from collections import defaultdict
from typing import List, Optional

import click
import datasets
import numpy as np
import torch
from safetensors.torch import save_file
from torch.utils.data import DataLoader
from transformers import AutoModel, AutoTokenizer, DefaultDataCollator

from mergekitty.architecture import _template_substitution, get_architecture_info
from mergekitty.common import ModelReference

logging.basicConfig(level=logging.INFO)

# set seed
torch.manual_seed(42)
np.random.seed(42)


def clean_name(name):
    return name.replace(".weight", "").replace("model.", "")


def parse_items(ctx, param, value):
    if value is not None:
        return [item.strip() for item in value.split(",")]


def remove_pads(attention_mask, feature_vector):
    if (
        len(feature_vector.shape) == 3
    ):  # Hidden states: (batch_size, seq_length, embedding_dim)
        # Expand mask to match the feature_vector dimensions and apply it
        expanded_mask = attention_mask.unsqueeze(-1)
        filtered_feature_vector = feature_vector * expanded_mask
    else:
        raise ValueError("Unsupported feature vector shape.")

    return filtered_feature_vector


def get_attention_output_hook(storage_dict, space_name, capture_input=True):
    """
    Returns a hook function that stores the output of the attention layer.
    """

    def hook(module, input, output):
        # NOTE: shape of input is [batch, seq_len, dim] and output is Tuple[(seq_len, dim),...]
        if capture_input:
            o = input[0].detach()
        else:
            o = output.detach()

        if space_name not in storage_dict:
            storage_dict[space_name] = o
        else:
            storage_dict[space_name] = torch.cat((storage_dict[space_name], o), dim=0)

    return hook


"""

What this script does:

It tries to map input/output spaces to activation maps

"""


@click.command("mergekitty-abm-extract-activations")
@click.argument("model-path", type=str)
@click.option(
    "--dataset", "-d", required=True, type=str, help="Dataset to use for activations"
)
@click.option("--out-path", "-o", required=True, type=str, help="Output model path")
@click.option("--batch-size", "-b", type=int, default=2, help="Batch size")
@click.option(
    "--dataset-size",
    "-s",
    type=int,
    default=None,
    help="Dataset size. If None, use full dataset",
)
@click.option(
    "--dataset-column", "-c", type=str, default="text", help="Dataset column to use"
)
@click.option(
    "--dataset-subset", "-u", type=str, default="eval", help="Dataset subset to use"
)
@click.option(
    "--chat-template/--no-chat-template",
    default=False,
    help="use Chat template for inference",
)
@click.option("--max-length", "-l", type=int, default=512, help="Max length")
@click.option("--dtype", type=str, default=None, help="Data type to convert weights to")
@click.option(
    "--device", type=str, default=None, help="device to compute the activations"
)
@click.option(
    "--ignore-spaces",
    "-i",
    type=str,
    default="",
    callback=parse_items,
    help="Spaces to ignore separated by comma. Example: up_${layer_index}",
)
def main(
    model_path: str,
    dataset: str,
    dataset_column: str,
    out_path: str,
    batch_size: int,
    max_length: int,
    dataset_size: Optional[int],
    dataset_subset: Optional[str],
    chat_template: Optional[bool],
    dtype: Optional[str],
    device: Optional[str],
    ignore_spaces: Optional[List[str]],
):
    # sorting out locations to hook into
    # we do this via the predefined json architecture definitions in mergekit

    model = ModelReference.model_validate(model_path)

    model_config = model.config()
    model_arch_info = get_architecture_info(model_config)

    _json = model_arch_info.definition

    residual_space = None

    weights = []
    for weight in _json.layer_templates.weights:
        if weight.is_kq:
            residual_space = weight.input_space
        weights.append(weight)

    if residual_space is None:
        raise ValueError("No residual space found")

    # ======================== Mapping spaces to weights ========================

    # just a list of connected components
    space_to_output_weight_templates = defaultdict(list)
    space_to_input_weight_templates = defaultdict(list)

    for layer_template in weights:
        if (
            not layer_template.input_space
            or layer_template.input_space in ignore_spaces
        ):
            continue
        space_to_output_weight_templates[layer_template.input_space].append(
            layer_template.name
        )

    for layer_template in weights:
        if (
            not layer_template.output_space
            or layer_template.output_space in ignore_spaces
        ):
            continue
        space_to_input_weight_templates[layer_template.output_space].append(
            layer_template.name
        )

    # remove the residual space from the input and output
    space_to_input_weight_templates.pop(residual_space, None)
    space_to_output_weight_templates.pop(residual_space, None)

    # NOTE: if space has input and output weights, remove one or the other because hooking
    # into both will result in duplicate activations
    to_remove = []
    for space, input_weights in space_to_input_weight_templates.items():
        if space in space_to_output_weight_templates:
            # if count of input weights and output weights is non zero, remove the space from space to output_weights
            if (
                len(input_weights) > 0
                and len(space_to_output_weight_templates[space]) > 0
            ):
                to_remove.append(space)

    # remove keys from output
    space_to_output_weight_templates = {
        k: v for k, v in space_to_output_weight_templates.items() if k not in to_remove
    }

    num_layers = model_arch_info.num_layers(model_config)

    space_to_input_weights = {}
    for k, v in space_to_input_weight_templates.items():
        for j in range(num_layers):

            def f(x):
                return _template_substitution(x, num_layers=num_layers, layer_idx=j)

            space_to_input_weights[f(k)] = [f(_v) for _v in v]

    space_to_output_weights = {}
    for k, v in space_to_output_weight_templates.items():
        for j in range(num_layers):

            def f(x):
                return _template_substitution(x, num_layers=num_layers, layer_idx=j)

            space_to_output_weights[f(k)] = [f(_v) for _v in v]

    # ================== Load model, tokenizer for inference and prepare dataset ==================

    model = AutoModel.from_pretrained(
        model_path, output_attentions=True, attn_implementation="eager"
    )
    tokenizer = AutoTokenizer.from_pretrained(model_path)

    if not tokenizer.pad_token:
        tokenizer.pad_token = tokenizer.eos_token

    tokenize_function = None
    if chat_template:
        logging.info("Using chat template for inference")

        def tokenize_function(x):
            return tokenizer.apply_chat_template(
                x,
                padding="longest",
                max_length=max_length,
                truncation=True,
                return_dict=True,
            )
    else:
        logging.info("Using default tokenizer (no chat template) for inference")

        def tokenize_function(x):
            return tokenizer(
                x,
                padding="longest",
                max_length=max_length,
                truncation=True,
            )

    model.eval()
    model.to(device)
    if dtype is not None:
        model = model.to(dtype=dtype)

    dataset = datasets.load_dataset(dataset)[dataset_subset]

    if dataset_size is not None:
        logging.info("Using dataset size %s", dataset_size)
        dataset = dataset.select(range(dataset_size))

    def tokenize(element):
        outputs = tokenize_function(element[dataset_column])
        return {
            "input_ids": outputs["input_ids"],
            "attention_mask": outputs["attention_mask"],
        }

    dataset = dataset.map(tokenize).select_columns(["input_ids", "attention_mask"])

    datasets_dataloader = DataLoader(
        dataset, batch_size=batch_size, shuffle=False, collate_fn=DefaultDataCollator()
    )

    feature_storage = {}
    storage_dict = {}

    # ================== Hooking into the model ==================

    # NOTE: if the capture input set to True seems confusing, a space's output is a weight recieving input from the space
    for k, v in space_to_output_weights.items():
        for weight in v:
            weight = clean_name(weight)
            model.get_submodule(weight).register_forward_hook(
                get_attention_output_hook(feature_storage, k, capture_input=True)
            )
    for k, v in space_to_input_weights.items():
        for weight in v:
            weight = clean_name(weight)
            model.get_submodule(weight).register_forward_hook(
                get_attention_output_hook(feature_storage, k, capture_input=False)
            )

    # ================== Inference ==================

    for batch in datasets_dataloader:
        with torch.no_grad():
            inputs = {k: v.to(device) for k, v in batch.items()}
            outputs = model(
                **inputs, output_hidden_states=True, output_attentions=False
            )

            # NOTE: https://huggingface.co/docs/transformers/en/main_classes/output#transformers.modeling_outputs.BaseModelOutput

            # Store attention masks
            attention_mask = inputs["attention_mask"]
            if "attention_mask" not in feature_storage:
                feature_storage["attention_mask"] = attention_mask.cpu().detach()
            else:
                feature_storage["attention_mask"] = torch.cat(
                    (feature_storage["attention_mask"], attention_mask.cpu().detach()),
                    dim=0,
                )

            hidden_states = [
                remove_pads(attention_mask, hidden_state)
                for hidden_state in outputs.hidden_states
            ]
            hidden_states = torch.stack(outputs.hidden_states, dim=1)

            if residual_space not in feature_storage:
                feature_storage[residual_space] = hidden_states
            else:
                feature_storage[residual_space] = torch.cat(
                    (feature_storage[residual_space], hidden_states), dim=0
                )

            for space_name, v in storage_dict.items():
                if space_name not in feature_storage:
                    feature_storage[space_name] = v
                else:
                    feature_storage[space_name] = torch.cat(
                        (feature_storage[space_name], v), dim=0
                    )

            storage_dict = {}

    # ================== Save activations/features ==================

    logging.info("Feature storage:")
    for k, v in feature_storage.items():
        if v is not None:
            logging.info(f"{k}: Shape: {v.shape}")

    abs_path = os.path.abspath(model_path)
    if os.path.exists(abs_path):
        model_path = abs_path

    model_path = model_path.replace("/", "_")

    # create output directory
    os.makedirs(out_path, exist_ok=True)

    save_file(
        feature_storage, os.path.join(out_path, f"{model_path}_features.safetensor")
    )


if __name__ == "__main__":
    main()



# Copyright (C) 2025 Arcee AI
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import logging
from typing import Dict, Optional, Tuple

import torch
import tqdm
import transformers

from mergekitty.architecture import WeightInfo
from mergekitty.common import ModelReference, dtype_from_name
from mergekitty.io import LazyTensorLoader, TensorWriter
from mergekitty.merge import MergeOptions
from mergekitty.moe.config import Expert, MoEMergeConfig


def initialize_io(
    config: MoEMergeConfig,
    out_path: str,
    merge_options: MergeOptions,
) -> Tuple[Dict[ModelReference, LazyTensorLoader], LazyTensorLoader, TensorWriter]:
    base_model = config.base_model
    loaders: Dict[ModelReference, LazyTensorLoader] = {}
    for model in tqdm.tqdm(
        [base_model] + [e.source_model for e in config.experts], desc="Warm up loaders"
    ):
        loaders[model] = model.lazy_loader(
            cache_dir=merge_options.transformers_cache,
            lazy_unpickle=merge_options.lazy_unpickle,
        )

    base_loader = loaders.get(base_model)
    writer = TensorWriter(
        out_path=out_path,
        max_shard_size=merge_options.out_shard_size,
        safe_serialization=merge_options.safe_serialization,
    )

    return loaders, base_loader, writer


def select_dtype(
    config: MoEMergeConfig, base_cfg: transformers.PretrainedConfig
) -> Optional[torch.dtype]:
    out_dtype = None
    if config.dtype:
        out_dtype = dtype_from_name(config.dtype)

    if out_dtype is None and base_cfg.torch_dtype:
        out_dtype = base_cfg.torch_dtype
        if isinstance(out_dtype, str):
            out_dtype = dtype_from_name(out_dtype)
    return out_dtype


def noise_and_scale(
    tensor: torch.Tensor, expert: Expert, is_residual: bool = False
) -> torch.Tensor:
    if expert.noise_scale is not None:
        noise = torch.randn_like(tensor) * expert.noise_scale
        tensor = tensor + noise
    if is_residual and expert.residual_scale is not None:
        tensor = tensor * expert.residual_scale
    return tensor


def copy_tensor_out(
    weight_info: WeightInfo,
    loader: LazyTensorLoader,
    writer: TensorWriter,
    expert: Optional[Expert] = None,
    is_residual: bool = False,
    output_name: Optional[str] = None,
    out_dtype: Optional[torch.dtype] = None,
    clone: bool = False,
):
    out_tensor_name = output_name or weight_info.name
    try:
        tensor = loader.get_tensor(weight_info.name, aliases=weight_info.aliases)
    except KeyError:
        tensor = None
    if tensor is None and not weight_info.optional:
        logging.error(f"Missing weight: {weight_info.name} / {out_tensor_name}")
        raise KeyError(out_tensor_name)

    if expert:
        tensor = noise_and_scale(tensor, expert, is_residual=is_residual)
    writer.save_tensor(
        out_tensor_name,
        tensor.to(dtype=out_dtype),
        clone=clone,
    )



# Partly Copyright (C) 2025 Arcee AI
# Partly Copyright (C) 2025-2026 Allura-org
import os
import tempfile
from typing import Callable, Optional, List, Union
import json

import pytest
from transformers import (
    AutoConfig,
    LlamaConfig,
    LlamaForCausalLM,
    PreTrainedTokenizerBase,
    LlamaTokenizerFast,
    Qwen2VLConfig,
    Qwen2VLForConditionalGeneration,
    Qwen3MoeConfig,
    Qwen3MoeForCausalLM,
)
import tokenizers

from mergekitty.architecture import get_architecture_info
from mergekitty.config import MergeConfiguration
from mergekitty.io.lazy_tensor_loader import LazyTensorLoader, ShardedTensorIndex
from mergekitty.merge import MergeOptions, run_merge


def make_tokenizer(
    vocab_size: int, added_tokens: List[Union[str, tokenizers.AddedToken]]
) -> PreTrainedTokenizerBase:
    tokens = ["<unk>", "<s>", "</s>"] + [f"_tok_{idx}" for idx in range(3, vocab_size)]
    tokens = tokens[:vocab_size]
    tok_data = {
        "version": "1.0",
        "model": {
            "type": "BPE",
            "vocab": dict(zip(tokens, range(vocab_size))),
            "merges": [],
        },
        "added_tokens": [],
    }
    tok = tokenizers.Tokenizer.from_str(json.dumps(tok_data))
    with tempfile.TemporaryDirectory() as p:
        tok_path = os.path.join(p, "tokenizer.json")
        tok.save(tok_path)
        res = LlamaTokenizerFast(tokenizer_file=tok_path)

    res.add_tokens(added_tokens)
    return res


def run_and_check_merge(
    config: MergeConfiguration,
    check_nan: bool = True,
    check_tensors: bool = True,
    validate: Optional[Callable[[str], None]] = None,
    index_json_name: Optional[str] = None,
):
    if index_json_name is None:
        index_json_name = "model.safetensors.index.json"

    with tempfile.TemporaryDirectory() as tmpdir:
        run_merge(config, out_path=tmpdir, options=MergeOptions())
        assert os.path.exists(os.path.join(tmpdir, index_json_name)), (
            "No index file for merge"
        )
        assert os.path.exists(os.path.join(tmpdir, "config.json")), (
            "No config json produced by merge"
        )

        if check_nan:
            # check for NaN in output
            loader = LazyTensorLoader.from_disk(tmpdir, lazy_unpickle=False)
            tp = loader.index.tensor_paths
            sorted_tensors = sorted(tp.keys(), key=lambda k: tp[k])
            for tensor_name in sorted_tensors:
                tensor = loader.get_tensor(tensor_name)
                has_nan = tensor.view(-1).isnan().any()
                assert not has_nan, "Output contains NaN"

        if check_tensors:
            config = AutoConfig.from_pretrained(tmpdir)
            arch_info = get_architecture_info(config)

            index = ShardedTensorIndex.from_disk(tmpdir)
            for weight_info in arch_info.all_weights(config):
                if weight_info.optional:
                    continue
                if weight_info.name not in index.tensor_paths and not any(
                    a in index.tensor_paths for a in weight_info.aliases
                ):
                    raise RuntimeError(f"Output missing tensor {weight_info.name}")

        if validate:
            validate(tmpdir)


def make_picollama(path: str, vocab_size: int = 64):
    cfg = LlamaConfig(
        vocab_size=vocab_size,
        hidden_size=32,
        intermediate_size=48,
        num_attention_heads=16,
        num_hidden_layers=2,
    )
    model = LlamaForCausalLM(cfg)
    model.save_pretrained(path, safe_serialization=True)
    return str(path)


def make_qwen3moe_picollama(path: str, vocab_size: int = 64):
    cfg = Qwen3MoeConfig(
        vocab_size=vocab_size,
        hidden_size=32,
        intermediate_size=48,
        num_hidden_layers=2,
        num_experts=2,
        num_experts_per_tok=2,
    )
    model = Qwen3MoeForCausalLM(cfg)
    model.save_pretrained(path, safe_serialization=True)
    return str(path)


def make_qwen2vl_picollama(path: str, vocab_size: int = 64):
    cfg = Qwen2VLConfig(
        text_config={
            "vocab_size": vocab_size,
            "hidden_size": 32,
            "intermediate_size": 64,
            "num_hidden_layers": 2,
            "num_attention_heads": 4,
            "num_key_value_heads": 4,
            "max_position_embeddings": 128,
            "rope_theta": 10000.0,
            "bos_token_id": 1,
            "eos_token_id": 2,
            "pad_token_id": 0,
        },
        vision_config={
            "depth": 2,
            "embed_dim": 32,
            "hidden_size": 32,
            "hidden_act": "gelu",
            "mlp_ratio": 2,
            "num_heads": 4,
            "in_channels": 3,
            "patch_size": 4,
            "spatial_merge_size": 2,
            "temporal_patch_size": 1,
        },
        image_token_id=3,
        video_token_id=4,
        vision_start_token_id=5,
        vision_end_token_id=6,
    )
    model = Qwen2VLForConditionalGeneration(cfg)
    model.config.architectures = ["Qwen2VLForConditionalGeneration"]
    model.save_pretrained(path, safe_serialization=True)
    return str(path)


def make_mistral3_picollama(path: str, vocab_size: int = 64):
    try:
        from transformers import Mistral3Config, Mistral3ForConditionalGeneration
    except ImportError:
        pytest.skip("transformers build does not include Mistral3")

    cfg = Mistral3Config(
        text_config={
            "model_type": "mistral",
            "vocab_size": vocab_size,
            "hidden_size": 32,
            "intermediate_size": 64,
            "num_hidden_layers": 2,
            "num_attention_heads": 4,
            "num_key_value_heads": 4,
            "max_position_embeddings": 128,
            "rope_theta": 10000.0,
            "bos_token_id": 1,
            "eos_token_id": 2,
            "pad_token_id": 0,
        },
        vision_config={
            "model_type": "pixtral",
            "hidden_size": 32,
            "intermediate_size": 64,
            "num_hidden_layers": 2,
            "num_attention_heads": 4,
            "num_channels": 3,
            "patch_size": 4,
            "image_size": 16,
            "hidden_act": "gelu",
        },
        image_token_index=3,
        spatial_merge_size=2,
        tie_word_embeddings=False,
    )
    model = Mistral3ForConditionalGeneration(cfg)
    model.config.architectures = ["Mistral3ForConditionalGeneration"]
    model.save_pretrained(path, safe_serialization=True)
    return str(path)



from typing import Dict, Optional

import pytest
from common import make_picollama, make_tokenizer, run_and_check_merge
from transformers import AutoConfig

from mergekitty.config import (
    InputModelDefinition,
    InputSliceDefinition,
    MergeConfiguration,
    OutputSliceDefinition,
    ParameterSetting,
)
from mergekitty.io import LazyTensorLoader


@pytest.fixture(scope="session")
def model_a(tmp_path_factory):
    model_path = make_picollama(tmp_path_factory.mktemp("model_a"))
    make_tokenizer(vocab_size=64, added_tokens=[]).save_pretrained(model_path)
    return model_path


@pytest.fixture(scope="session")
def model_b(tmp_path_factory):
    model_path = make_picollama(tmp_path_factory.mktemp("model_b"))
    make_tokenizer(vocab_size=64, added_tokens=[]).save_pretrained(model_path)
    return model_path


@pytest.fixture(scope="session")
def model_c(tmp_path_factory):
    model_path = make_picollama(tmp_path_factory.mktemp("model_c"))
    make_tokenizer(vocab_size=64, added_tokens=[]).save_pretrained(model_path)
    return model_path


class TestBasicMerges:
    def test_gpt2_copy(self):
        config = MergeConfiguration(
            merge_method="passthrough",
            models=[InputModelDefinition(model="gpt2")],
            dtype="bfloat16",
        )
        run_and_check_merge(config)

    def test_gpt2_stack(self):
        config = MergeConfiguration(
            merge_method="passthrough",
            slices=[
                OutputSliceDefinition(
                    sources=[InputSliceDefinition(model="gpt2", layer_range=[0, 12])]
                )
            ]
            * 2,
            dtype="bfloat16",
        )

        def _check_config_layers(p: str):
            config = AutoConfig.from_pretrained(p)
            assert config.n_layer == 24

        run_and_check_merge(config, validate=_check_config_layers)

    def test_passthrough_scale(self, model_a):
        config = MergeConfiguration(
            merge_method="passthrough",
            models=[
                InputModelDefinition(
                    model=model_a,
                    parameters={
                        "scale": [
                            {"filter": "o_proj", "value": 0},
                            {"value": 1},
                        ]
                    },
                )
            ],
        )

        def _check_o_proj(p: str):
            loader = LazyTensorLoader.from_disk(p)
            saw_any = False
            for name in loader.index.tensor_paths:
                if "o_proj" in name:
                    param = loader.get_tensor(name)
                    assert (param == 0).all()
                    saw_any = True
                elif "lm_head" in name:
                    param = loader.get_tensor(name)
                    assert param.count_nonzero() > 0

            assert saw_any, "No o_proj parameters found"

        run_and_check_merge(config, validate=_check_o_proj)

    def test_linear_merge(self, model_a, model_b):
        config = self.two_model_config(model_a, model_b, merge_method="linear")
        run_and_check_merge(config)

    def test_slerp_merge(self, model_a, model_b):
        config = self.two_model_config(
            model_a, model_b, merge_method="slerp", base_model=model_a
        )
        config.parameters = {"t": 0.35}
        run_and_check_merge(config)

    def test_nearswap_merge(self, model_a, model_b):
        config = self.two_model_config(
            model_a, model_b, merge_method="nearswap", base_model=model_a
        )
        config.parameters = {"t": 0.0001}
        run_and_check_merge(config)

    def test_nuslerp_merges(self, model_a, model_b, model_c):
        for base_model in [None, model_c]:
            for row_wise in [False, True]:
                for flatten in [False, True]:
                    print(
                        f"Testing nuslerp with row_wise={row_wise}, flatten={flatten}, base_model={base_model}"
                    )
                    run_and_check_merge(
                        self.two_model_config(
                            model_a,
                            model_b,
                            merge_method="slerp",
                            base_model=base_model,
                            params={
                                "nuslerp_row_wise": row_wise,
                                "nuslerp_flatten": flatten,
                            },
                        )
                    )

        # test weights that sum to zero
        config = self.two_model_config(
            model_a,
            model_b,
            merge_method="slerp",
            base_model=model_c,
            params={"nuslerp_row_wise": False, "nuslerp_flatten": False},
        )
        config.models[0].parameters["weight"] = -0.5
        config.models[1].parameters["weight"] = 0.5
        run_and_check_merge(config)

    def test_task_arithmetic_merge(self, model_a, model_b, model_c):
        config = self.two_model_config(
            model_a, model_b, merge_method="task_arithmetic", base_model=model_c
        )
        run_and_check_merge(config)

    def test_breadcrumbs_merge(self, model_a, model_b, model_c):
        config = self.two_model_config(
            model_a, model_b, merge_method="breadcrumbs", base_model=model_c
        )
        run_and_check_merge(config)

    def test_ties_merge(self, model_a, model_b, model_c):
        config = self.two_model_config(
            model_a,
            model_b,
            merge_method="ties",
            base_model=model_c,
            params={"density": 0.3},
        )
        run_and_check_merge(config)

    def test_multislerp_merge(self, model_a, model_b, model_c):
        config = self.two_model_config(
            model_a,
            model_b,
            merge_method="multislerp",
            base_model=model_c,
        )
        run_and_check_merge(config)

    def test_dare_ties_merge(self, model_a, model_b, model_c):
        config = self.two_model_config(
            model_a,
            model_b,
            merge_method="dare_ties",
            base_model=model_c,
            params={"density": 0.66},
        )
        run_and_check_merge(config)

    def test_model_stock_merge(self, model_a, model_b, model_c):
        config = self.two_model_config(
            model_b, model_c, merge_method="model_stock", base_model=model_a
        )
        run_and_check_merge(config)

    def test_model_stock_filterwise_merge(self, model_a, model_b, model_c):
        config = self.two_model_config(
            model_b,
            model_c,
            merge_method="model_stock",
            base_model=model_a,
            params={"filter_wise": True},
        )
        run_and_check_merge(config)

    def two_model_config(
        self,
        model_a,
        model_b,
        merge_method: str,
        base_model: Optional[str] = None,
        params: Optional[Dict[str, ParameterSetting]] = None,
    ):
        config = MergeConfiguration(
            merge_method=merge_method,
            base_model=base_model,
            models=[
                InputModelDefinition(
                    model=model_a,
                    parameters={"weight": 0.6},
                ),
                InputModelDefinition(
                    model=model_b,
                    parameters={"weight": 0.4},
                ),
            ],
            dtype="bfloat16",
            parameters=params,
        )

        return config



# Copyright (C) 2025 Arcee AI
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

from abc import ABC, abstractmethod
from typing import List, Optional

import torch

from mergekitty.moe.config import MoEMergeConfig
from mergekitty.options import MergeOptions


class MoEOutputArchitecture(ABC):
    @abstractmethod
    def name(self) -> str:
        """Return a human-readable name for the architecture."""
        pass

    @abstractmethod
    def supports_config(
        self,
        config: MoEMergeConfig,
        explain: bool = False,
        trust_remote_code: bool = False,
    ) -> bool:
        """Return whether this architecture supports the given config.

        If `explain` is True, log an explanation of why the config is not supported."""
        pass

    @abstractmethod
    def write_model(
        self,
        out_path: str,
        config: MoEMergeConfig,
        merge_options: MergeOptions,
        router_weights: List[torch.Tensor],
        shared_router_weights: Optional[List[torch.Tensor]] = None,
    ):
        """Write the config and tensors for the output MoE to the given path."""
        pass



# Copyright (C) 2025 Arcee AI
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.


import logging
from typing import List

import torch

from mergekitty.architecture import WeightInfo


def rectify_embed_sizes(weight_info: WeightInfo, tensors: List[torch.Tensor]):
    # TODO: use arch_info.embed_weights() instead
    if weight_info.is_embed and all(len(t.shape) == 2 for t in tensors):
        # special case - if lm_head.weight or embed_tokens.weight have a size
        # mismatch, take the largest common submatrix of all of them
        if take_common_submatrix(tensors):
            logging.warning(
                f"Using common submatrix of size {tensors[0].shape} for {weight_info.name}"
            )


def take_common_submatrix(tensors: List[torch.Tensor]) -> bool:
    min_size = [None, None]
    for t in tensors:
        for idx in range(2):
            if min_size[idx] is None or t.shape[idx] < min_size[idx]:
                min_size[idx] = t.shape[idx]

    if not all(t.shape == torch.Size(min_size) for t in tensors):
        for idx in range(len(tensors)):
            tensors[idx] = tensors[idx][: min_size[0], : min_size[1]]
        return True
    return False



# Partly Copyright (C) 2025 Arcee AI
# Partly Copyright (C) 2025-2026 Allura-org
from typing import Optional

import pytest
from common import make_picollama, make_tokenizer, run_and_check_merge
from transformers import AutoTokenizer

from mergekitty.config import InputModelDefinition, MergeConfiguration


def check_chat_template(model_path: str, needle: Optional[str] = None):
    tokenizer = AutoTokenizer.from_pretrained(model_path)
    if needle is None:
        assert not tokenizer.chat_template, "Expected no chat template"
        return
    assert tokenizer.chat_template and needle in tokenizer.chat_template, (
        f"Expected chat template to contain {needle}"
    )


@pytest.fixture(scope="session")
def model_base(tmp_path_factory):
    model_path = make_picollama(tmp_path_factory.mktemp("model_base"), vocab_size=64)
    make_tokenizer(vocab_size=64, added_tokens=[]).save_pretrained(model_path)
    return model_path


@pytest.fixture(scope="session")
def model_b(tmp_path_factory):
    model_path = make_picollama(tmp_path_factory.mktemp("model_b"), vocab_size=64)
    make_tokenizer(vocab_size=64, added_tokens=[]).save_pretrained(model_path)
    return model_path


class TestChatTemplate:
    def test_template_chatml(self, model_base, model_b):
        config = MergeConfiguration(
            merge_method="linear",
            models=[
                InputModelDefinition(model=model_base, parameters={"weight": 0.5}),
                InputModelDefinition(model=model_b, parameters={"weight": 0.5}),
            ],
            base_model=model_base,
            dtype="bfloat16",
            chat_template="chatml",
        )
        run_and_check_merge(
            config,
            validate=lambda p: check_chat_template(p, "<|im_start|>"),
        )

    def test_template_literal_jinja(self, model_base, model_b):
        config = MergeConfiguration(
            merge_method="linear",
            models=[
                InputModelDefinition(model=model_base, parameters={"weight": 0.5}),
                InputModelDefinition(model=model_b, parameters={"weight": 0.5}),
            ],
            base_model=model_base,
            dtype="bfloat16",
            chat_template="{{messages[0]['content']}}",
        )
        run_and_check_merge(
            config,
            validate=lambda p: check_chat_template(p, "{{messages[0]['content']}}"),
        )



# Copyright (C) 2025-2026 Allura-org
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import logging
import os
from dataclasses import dataclass, field
from typing import Optional

import click
import transformers

from mergekitty.architecture import WeightInfo, get_architecture_info
from mergekitty.card import generate_card_merged_lora
from mergekitty.common import AdapterReference, ModelPath, ModelReference
from mergekitty.custom_task import MergeLoraTask
from mergekitty.executor import build_executor
from mergekitty.io.tasks import (
    FinalizeModel,
    LoadTensor,
    LoaderCache,
    SaveTensor,
    TensorWriterTask,
)
from mergekitty.options import MergeOptions, add_merge_options

_ADAPTER_PREFIX = "base_model.model."
_FULL_WEIGHT_SUFFIXES = (
    ".modules_to_save.weight",
    ".original_module.weight",
    ".base_layer.weight",
    ".weight",
)
_FULL_WEIGHT_PRIORITY = {
    ".weight": 0,
    ".modules_to_save.weight": 1,
    ".original_module.weight": 2,
    ".base_layer.weight": 3,
}


@dataclass
class LoraModuleSpec:
    module_name: str
    tensor_a: Optional[str] = None
    tensor_b: Optional[str] = None
    base_tensor: Optional[str] = None
    is_embedding: bool = False

    def is_complete(self) -> bool:
        return self.tensor_a is not None and self.tensor_b is not None


@dataclass
class AdapterLayout:
    lora_modules: dict[str, LoraModuleSpec] = field(default_factory=dict)
    full_modules: dict[str, str] = field(default_factory=dict)


@dataclass
class PlanResults:
    tasks: list
    base_vocab_size: int
    merged_vocab_size: int


def _module_name_from_tensor(tensor_name: str, suffix: str) -> str:
    stripped = tensor_name[len(_ADAPTER_PREFIX) :]
    return stripped[: -len(suffix)]


def _parse_adapter_layout(adapter_ref: AdapterReference) -> AdapterLayout:
    loader = LoaderCache().get(adapter_ref)
    layout = AdapterLayout()
    full_weight_candidates: dict[str, tuple[int, str]] = {}

    for tensor_name in loader.index.tensor_paths:
        if not tensor_name.startswith(_ADAPTER_PREFIX):
            continue

        if tensor_name.endswith(".lora_A.weight"):
            module_name = _module_name_from_tensor(tensor_name, ".lora_A.weight")
            layout.lora_modules.setdefault(
                module_name, LoraModuleSpec(module_name=module_name)
            ).tensor_a = tensor_name
            continue

        if tensor_name.endswith(".lora_B.weight"):
            module_name = _module_name_from_tensor(tensor_name, ".lora_B.weight")
            layout.lora_modules.setdefault(
                module_name, LoraModuleSpec(module_name=module_name)
            ).tensor_b = tensor_name
            continue

        if tensor_name.endswith(".lora_embedding_A"):
            module_name = _module_name_from_tensor(tensor_name, ".lora_embedding_A")
            spec = layout.lora_modules.setdefault(
                module_name, LoraModuleSpec(module_name=module_name)
            )
            spec.tensor_a = tensor_name
            spec.is_embedding = True
            continue

        if tensor_name.endswith(".lora_embedding_B"):
            module_name = _module_name_from_tensor(tensor_name, ".lora_embedding_B")
            spec = layout.lora_modules.setdefault(
                module_name, LoraModuleSpec(module_name=module_name)
            )
            spec.tensor_b = tensor_name
            spec.is_embedding = True
            continue

        for suffix in _FULL_WEIGHT_SUFFIXES:
            if not tensor_name.endswith(suffix):
                continue

            module_name = _module_name_from_tensor(tensor_name, suffix)
            if suffix == ".base_layer.weight":
                layout.lora_modules.setdefault(
                    module_name, LoraModuleSpec(module_name=module_name)
                ).base_tensor = tensor_name

            current = full_weight_candidates.get(module_name)
            priority = _FULL_WEIGHT_PRIORITY[suffix]
            if current is None or priority < current[0]:
                full_weight_candidates[module_name] = (priority, tensor_name)
            break

    incomplete = [
        module_name
        for module_name, spec in layout.lora_modules.items()
        if (spec.tensor_a is None) != (spec.tensor_b is None)
    ]
    if incomplete:
        raise RuntimeError(
            "Adapter contains incomplete LoRA tensor pairs for modules: "
            + ", ".join(sorted(incomplete))
        )

    for module_name, (_priority, tensor_name) in full_weight_candidates.items():
        spec = layout.lora_modules.get(module_name)
        if spec and spec.is_complete() and spec.base_tensor == tensor_name:
            continue
        layout.full_modules[module_name] = tensor_name

    return layout


def _resolve_weight_tensor_name(loader, weight: WeightInfo) -> Optional[str]:
    names = [weight.name] + list(weight.aliases or ()) + list(weight.tied_names or ())
    for name in names:
        if name in loader.index.tensor_paths:
            return name
    return None


def _weight_module_candidates(weight: WeightInfo) -> list[str]:
    names = [weight.name] + list(weight.aliases or ()) + list(weight.tied_names or ())
    res = []
    for name in names:
        if name.endswith(".weight"):
            res.append(name[: -len(".weight")])
    return res


def _matching_module_name(weight: WeightInfo, modules: set[str]) -> Optional[str]:
    for module_name in _weight_module_candidates(weight):
        if module_name in modules:
            return module_name
    return None


def _module_alpha(adapter_config, module_name: str) -> float:
    alpha_pattern = getattr(adapter_config, "alpha_pattern", {}) or {}
    return float(alpha_pattern.get(module_name, adapter_config.lora_alpha))


def _validate_supported_adapter(adapter_config) -> None:
    if getattr(adapter_config, "peft_type", None) != "LORA":
        raise RuntimeError(
            "mergekitty-merge-lora currently supports only LoRA adapters"
        )
    if getattr(adapter_config, "bias", "none") != "none":
        raise RuntimeError("LoRA adapters with bias terms are not yet supported")
    if getattr(adapter_config, "use_dora", False):
        raise RuntimeError("DoRA adapters are not yet supported")


def _load_tokenizer(
    base_model_ref: ModelReference,
    adapter_ref: AdapterReference,
    trust_remote_code: bool,
):
    for source_name, path, revision in [
        ("adapter", adapter_ref.adapter.path, adapter_ref.adapter.revision),
        ("base", base_model_ref.model.path, base_model_ref.model.revision),
    ]:
        try:
            tokenizer = transformers.AutoTokenizer.from_pretrained(
                path,
                revision=revision,
                trust_remote_code=trust_remote_code,
            )
            return tokenizer, source_name
        except Exception as exc:
            logging.info(
                "Unable to load tokenizer from %s source %s",
                source_name,
                path,
                exc_info=exc,
            )

    return None, None


def _reconstruct_invocation(
    base_model: str,
    adapter_model: str,
    options: MergeOptions,
) -> str:
    defaults = MergeOptions()
    parts = [
        "mergekitty-merge-lora",
        base_model,
        adapter_model,
        "OUTPUT_PATH",
    ]

    for field_name, info in MergeOptions.model_fields.items():
        value = getattr(options, field_name)
        default = getattr(defaults, field_name)
        if value == default:
            continue

        option_name = field_name.replace("_", "-")
        if info.annotation is bool:
            parts.append(f"--{option_name}" if value else f"--no-{option_name}")
        else:
            parts.extend([f"--{option_name}", str(value)])

    return " ".join(parts)


def plan_merge(
    base_model_ref: ModelReference,
    adapter_ref: AdapterReference,
    output_path: str,
    options: MergeOptions,
) -> PlanResults:
    base_config = base_model_ref.config(trust_remote_code=options.trust_remote_code)
    arch_info = get_architecture_info(base_config)
    adapter_config = adapter_ref.config(trust_remote_code=options.trust_remote_code)
    _validate_supported_adapter(adapter_config)
    layout = _parse_adapter_layout(adapter_ref)

    base_loader = LoaderCache().get(base_model_ref)
    adapter_loader = LoaderCache().get(adapter_ref)

    writer_task = TensorWriterTask(
        out_path=output_path,
        max_shard_size=options.out_shard_size,
        safe_serialization=options.safe_serialization,
    )

    save_tasks = []
    used_modules = set()

    base_vocab_size = getattr(base_config, "vocab_size", 0)
    merged_vocab_size = base_vocab_size

    for weight in arch_info.all_weights(base_config):
        device = "cuda" if options.read_to_gpu else None
        lora_module_name = _matching_module_name(
            weight,
            {
                module_name
                for module_name, spec in layout.lora_modules.items()
                if spec.is_complete()
            },
        )
        full_module_name = _matching_module_name(weight, set(layout.full_modules))

        if lora_module_name:
            spec = layout.lora_modules[lora_module_name]
            base_tensor_task: LoadTensor
            if spec.base_tensor:
                base_tensor_task = LoadTensor(
                    model=adapter_ref,
                    tensor=spec.base_tensor,
                    device=device,
                    optional=weight.optional,
                )
            else:
                base_tensor_task = LoadTensor(
                    model=base_model_ref,
                    tensor=weight.name,
                    device=device,
                    optional=weight.optional,
                    aliases=weight.aliases,
                    tied_names=weight.tied_names,
                )

            tensor_task = MergeLoraTask(
                base_tensor=base_tensor_task,
                adapter_tensor_a=LoadTensor(
                    model=adapter_ref,
                    tensor=spec.tensor_a,
                    device=device,
                ),
                adapter_tensor_b=LoadTensor(
                    model=adapter_ref,
                    tensor=spec.tensor_b,
                    device=device,
                ),
                lora_alpha=_module_alpha(adapter_config, lora_module_name),
                use_rslora=getattr(adapter_config, "use_rslora", False),
                fan_in_fan_out=getattr(adapter_config, "fan_in_fan_out", False),
                is_embedding=spec.is_embedding,
            )
            used_modules.add(lora_module_name)
        elif full_module_name:
            tensor_task = LoadTensor(
                model=adapter_ref,
                tensor=layout.full_modules[full_module_name],
                device=device,
                optional=weight.optional,
            )
            used_modules.add(full_module_name)
        else:
            tensor_task = LoadTensor(
                model=base_model_ref,
                tensor=weight.name,
                device=device,
                optional=weight.optional,
                aliases=weight.aliases,
                tied_names=weight.tied_names,
            )

        save_tasks.append(
            SaveTensor(
                tensor_name=weight.name,
                tensor_task=tensor_task,
                writer_task=writer_task,
                clone=options.clone_tensors,
                optional=weight.optional,
                dtype=weight.force_dtype,
            )
        )

        if weight.is_embed:
            base_tensor_name = _resolve_weight_tensor_name(base_loader, weight)
            if base_tensor_name:
                base_vocab_size = max(
                    base_vocab_size,
                    base_loader.get_tensor(base_tensor_name).shape[0],
                )

            if lora_module_name and layout.lora_modules[lora_module_name].base_tensor:
                merged_vocab_size = max(
                    merged_vocab_size,
                    adapter_loader.get_tensor(
                        layout.lora_modules[lora_module_name].base_tensor
                    ).shape[0],
                )
            elif full_module_name:
                merged_vocab_size = max(
                    merged_vocab_size,
                    adapter_loader.get_tensor(
                        layout.full_modules[full_module_name]
                    ).shape[0],
                )
            elif base_tensor_name:
                merged_vocab_size = max(
                    merged_vocab_size,
                    base_loader.get_tensor(base_tensor_name).shape[0],
                )

    expected_modules = {
        module_name
        for module_name, spec in layout.lora_modules.items()
        if spec.is_complete()
    } | set(layout.full_modules)
    unused_modules = expected_modules - used_modules
    if unused_modules:
        raise RuntimeError(
            "Adapter contains modules that do not match output weights: "
            + ", ".join(sorted(unused_modules))
        )

    finalize_task = FinalizeModel(
        tensor_save_tasks=tuple(save_tasks),
        writer_task=writer_task,
    )

    return PlanResults(
        tasks=save_tasks + [finalize_task],
        base_vocab_size=base_vocab_size,
        merged_vocab_size=merged_vocab_size,
    )


def run_merge_lora(
    base_model: str,
    adapter_model: str,
    output_path: str,
    merge_options: MergeOptions,
) -> None:
    logging.basicConfig(
        level=logging.WARNING if merge_options.quiet else logging.INFO,
        format="%(asctime)s - %(levelname)s - %(message)s",
        force=True,
    )

    base_model_ref = ModelReference.model_validate(base_model)
    if base_model_ref.lora is not None:
        raise RuntimeError("BASE_MODEL must not include an inline LoRA reference")

    adapter_ref = AdapterReference(adapter=ModelPath.model_validate(adapter_model))

    loader_cache = LoaderCache()
    loader_cache.setup(merge_options)

    adapter_config = adapter_ref.config(
        trust_remote_code=merge_options.trust_remote_code
    )
    _validate_supported_adapter(adapter_config)

    if (
        getattr(adapter_config, "base_model_name_or_path", None)
        and adapter_config.base_model_name_or_path != base_model_ref.model.path
    ):
        logging.warning(
            "Adapter was created for %s but merge is using %s",
            adapter_config.base_model_name_or_path,
            base_model_ref.model.path,
        )

    os.makedirs(output_path, exist_ok=True)

    plan = plan_merge(
        base_model_ref=base_model_ref,
        adapter_ref=adapter_ref,
        output_path=output_path,
        options=merge_options,
    )

    exec = build_executor(
        tasks=plan.tasks,
        math_device="cuda" if merge_options.cuda else "cpu",
        storage_device="cuda" if merge_options.low_cpu_memory else "cpu",
        executor=merge_options.executor,
    )
    exec.execute()

    cfg_out = base_model_ref.config(trust_remote_code=merge_options.trust_remote_code)
    cfg_out.vocab_size = plan.merged_vocab_size
    cfg_out.save_pretrained(output_path)

    tokenizer = None
    if merge_options.copy_tokenizer:
        tokenizer, tokenizer_source = _load_tokenizer(
            base_model_ref=base_model_ref,
            adapter_ref=adapter_ref,
            trust_remote_code=merge_options.trust_remote_code,
        )
        if tokenizer is None:
            logging.warning(
                "No tokenizer could be loaded from the adapter or base model"
            )
        else:
            logging.info("Saving %s tokenizer", tokenizer_source)
            tokenizer.save_pretrained(output_path, safe_serialization=True)
            if len(tokenizer) != plan.merged_vocab_size:
                logging.warning(
                    "Tokenizer size %s does not match merged vocab size %s",
                    len(tokenizer),
                    plan.merged_vocab_size,
                )

    if merge_options.write_model_card:
        invocation = _reconstruct_invocation(
            base_model=base_model,
            adapter_model=adapter_model,
            options=merge_options,
        )
        card_md = generate_card_merged_lora(
            base_model_ref=base_model_ref,
            adapter_ref=adapter_ref,
            invocation=invocation,
            extended=plan.merged_vocab_size > plan.base_vocab_size,
            vocab_size=plan.merged_vocab_size,
            name=os.path.basename(output_path),
        )
        with open(os.path.join(output_path, "README.md"), "w", encoding="utf-8") as fp:
            fp.write(card_md)


@click.command("mergekitty-merge-lora")
@click.argument("base_model", type=str)
@click.argument("adapter_model", type=str)
@click.argument("output_path", type=str)
@add_merge_options
def main(
    base_model: str,
    adapter_model: str,
    output_path: str,
    merge_options: MergeOptions,
) -> None:
    """Merge a LoRA adapter with a base model to create a new model."""

    run_merge_lora(
        base_model=base_model,
        adapter_model=adapter_model,
        output_path=output_path,
        merge_options=merge_options,
    )


if __name__ == "__main__":
    main()



# Copyright (C) 2025 Arcee AI
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

from mergekitty.io.lazy_tensor_loader import (
    LazyTensorLoader,
    ShardedTensorIndex,
    ShardInfo,
)
from mergekitty.io.tensor_writer import TensorWriter

__all__ = [
    "LazyTensorLoader",
    "ShardedTensorIndex",
    "ShardInfo",
    "TensorWriter",
]



# Copyright (C) 2025 Arcee AI
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import json
import logging
import os
import os.path
import threading
from dataclasses import dataclass
from typing import Dict, List, Optional

import safetensors
import safetensors.torch
import torch
from torch import Tensor

from mergekitty.io.loader import TensorLoader


@dataclass
class ShardInfo:
    filename: str
    contained_keys: List[str]


@dataclass
class ShardedTensorIndex:
    base_path: str
    is_safetensors: bool
    tensor_paths: Dict[str, str]
    shards: List[ShardInfo]

    @classmethod
    def from_disk(cls, base_path: str) -> "ShardedTensorIndex":
        model_path = None
        for model_file_name in [
            "model.safetensors",
            "adapter_model.safetensors",
            "adapter_model.bin",
            "pytorch_model.bin",
        ]:
            candidate_path = os.path.join(base_path, model_file_name)
            if os.path.exists(candidate_path) or os.path.exists(
                candidate_path + ".index.json"
            ):
                model_path = candidate_path
                break

        if not model_path:
            raise RuntimeError(f"Unable to find model files at {base_path}")

        is_safetensors = model_path.endswith(".safetensors")
        tensor_paths = None
        shards = []

        if os.path.exists(model_path + ".index.json"):
            # shared model - parse index
            with open(model_path + ".index.json", "r") as fd:
                weight_map = json.load(fd)["weight_map"]
            tensor_paths = weight_map

            shard_names = list(sorted(set(tensor_paths[e] for e in tensor_paths)))
            for shard_name in shard_names:
                info = ShardInfo(
                    shard_name,
                    [key for key in tensor_paths if tensor_paths[key] == shard_name],
                )
                shards.append(info)

        elif os.path.exists(model_path):
            shard_name = os.path.basename(model_path)

            # get list of tensors contained in single-file checkpoint
            if model_path.lower().endswith(".safetensors"):
                with safetensors.safe_open(model_path, framework="pt") as st:
                    tensor_paths = {key: shard_name for key in st.keys()}
            else:
                # this is ugly but not much else can be done
                shard = torch.load(model_path, map_location="meta")
                if "state_dict" in shard:
                    shard = shard["state_dict"]

                tensor_paths = {key: shard_name for key in shard}

            shards.append(
                ShardInfo(os.path.basename(model_path), list(tensor_paths.keys()))
            )

        return ShardedTensorIndex(
            base_path=base_path,
            is_safetensors=is_safetensors,
            tensor_paths=tensor_paths,
            shards=shards,
        )


@dataclass
class _ThreadShardState:
    shard_file: Optional[str] = None
    shard: Optional[TensorLoader] = None


class LazyTensorLoader:
    index: ShardedTensorIndex
    lazy_unpickle: bool

    def __init__(self, index: ShardedTensorIndex, lazy_unpickle: bool = True):
        self.index = index
        self.lazy_unpickle = lazy_unpickle
        self._thread_state = threading.local()

    def _get_thread_state(self) -> _ThreadShardState:
        state = getattr(self._thread_state, "value", None)
        if state is None:
            state = _ThreadShardState()
            self._thread_state.value = state
        return state

    def get_tensor(
        self,
        key: str,
        device: str = "cpu",
        aliases: Optional[List[str]] = None,
        raise_on_missing: bool = True,
    ) -> Optional[Tensor]:
        if aliases and key not in self.index.tensor_paths:
            for alias in aliases:
                if alias in self.index.tensor_paths:
                    key = alias
                    break

        if key not in self.index.tensor_paths:
            if raise_on_missing:
                raise KeyError(key)
            return None

        shard_file = self.index.tensor_paths[key]
        state = self._get_thread_state()
        if state.shard is None or state.shard_file != shard_file:
            shard_full_path = os.path.join(self.index.base_path, shard_file)
            logging.debug(f"Opening shard {shard_full_path}")
            state.shard = TensorLoader.get(
                shard_full_path,
                use_lazy_unpickle=self.lazy_unpickle,
                device=device,
            )
            state.shard_file = shard_file

        return state.shard.get_tensor(key).to(device)

    def flush(self):
        self._thread_state = threading.local()

    @classmethod
    def from_disk(
        cls, base_path: str, lazy_unpickle: bool = True
    ) -> "LazyTensorLoader":
        return LazyTensorLoader(ShardedTensorIndex.from_disk(base_path), lazy_unpickle)



import os
import sys
from collections import defaultdict

import click
import numpy as np
import safetensors.torch
import scipy
import torch

from mergekitty.architecture import _template_substitution, get_architecture_info
from mergekitty.common import ModelReference


def calc_correlation_matrix(feats):
    feats = feats.view(-1, feats.shape[-1])

    return torch.corrcoef(feats.T)


def match_tensors_permute(
    absval=False,
    correlation_matrix=None,
):
    """
    This function is adapted from ZipIt! (https://github.com/gstoica27/ZipIt)
    """

    Om = correlation_matrix.shape[0] // 2
    device = correlation_matrix.device

    mats = [torch.eye(Om, device=device)]

    corr_submatrix = correlation_matrix[:Om, Om:].cpu().numpy()
    if absval:
        corr_submatrix = np.absolute(corr_submatrix)
    _, col_ind = scipy.optimize.linear_sum_assignment(corr_submatrix, maximize=True)

    new_mat = torch.eye(Om, device=device)[torch.tensor(col_ind).long().to(device)]
    mats.append(new_mat.T)

    unmerge_mats = mats

    unmerge = torch.cat(unmerge_mats, dim=0)

    merge = torch.cat(mats, dim=0)
    merge = merge / (merge.sum(dim=0, keepdim=True) + 1e-5)

    return merge.T, unmerge


def match_tensors_permute_MHA(
    n_heads=32,
    absval=False,
    correlation_matrix=None,
):
    """
    Handles different head permutations in attention.
    Modified version of the function here: https://github.com/nverma1/merging-text-transformers/blob/main/matching_functions.py#L76
    """

    Om = correlation_matrix.shape[0] // 2
    device = correlation_matrix.device
    query_size = Om // n_heads

    mats = [torch.eye(Om, device=device)]
    head_perms = []

    costs = np.ones((n_heads, n_heads)) * -sys.maxsize

    col_inds_storage = defaultdict(lambda: defaultdict(int))

    for j in range(n_heads):
        for k in range(n_heads):
            head1_idx = [query_size * j, query_size * (j + 1)]
            head2_idx = [query_size * k, query_size * (k + 1)]

            corr_submatrix = (
                correlation_matrix[
                    head1_idx[0] : head1_idx[1],
                    (Om + head2_idx[0]) : (Om + head2_idx[1]),
                ]
                .cpu()
                .numpy()
            )
            if absval:
                corr_submatrix = np.absolute(corr_submatrix)

            # compute perm for head j & head k
            row_ind, col_ind = scipy.optimize.linear_sum_assignment(
                corr_submatrix, maximize=True
            )

            costs[j, k] = corr_submatrix[row_ind, col_ind].sum()

            col_inds_storage[j][k] = col_ind

    outer_row_ind, outer_col_ind = scipy.optimize.linear_sum_assignment(
        costs, maximize=True
    )

    for j in range(n_heads):
        head_1 = outer_row_ind[j]
        head_2 = outer_col_ind[j]

        head_perm = col_inds_storage[head_1][head_2]
        head_perms.append(torch.tensor(head_perm + query_size * head_2))

    new_mat = torch.eye(Om, device=device)[
        torch.cat(head_perms).clone().detach().long().to(device)
    ]
    mats.append(new_mat.T)

    unmerge_mats = mats

    unmerge = torch.cat(unmerge_mats, dim=0)
    merge = torch.cat(mats, dim=0)
    merge = merge / (merge.sum(dim=0, keepdim=True) + 1e-5)

    return merge.T, unmerge


@click.command("mergekitty-abm-extract-permutations")
@click.argument("model1-ft", type=str, required=True)
@click.argument("model2-ft", type=str, required=True)
@click.option("--model_path", type=str, required=True, help="Model information")
@click.option(
    "--out_path", required=True, type=str, help="Output path for metric tensors"
)
@click.option(
    "--absval/--no-absval",
    required=False,
    default=False,
    help="Use absolute value on correlation matrices/submatrices while calculating merge/unmerge matrices",
)
@click.option(
    "--device",
    "-d",
    type=str,
    default="cpu",
    help="Device to compute on (default: cpu)",
)
def main(model1_ft, model2_ft, model_path, out_path, absval, device):
    os.makedirs(out_path, exist_ok=True)

    model = ModelReference.model_validate(model_path)

    model_config = model.config()

    model_arch_info = get_architecture_info(model_config)

    _json = model_arch_info.definition

    residual_space = None
    kq_space = None
    v_space = None

    # extract the residual, attention related spaces
    for weight in _json.layer_templates.weights:
        if weight.is_kq:
            kq_space = weight.output_space
            residual_space = weight.input_space
            continue

        # assuming order is observed
        if (
            not weight.is_kq
            and weight.head_split
            and (weight.input_space == residual_space)
        ):
            v_space = weight.output_space
            continue

    num_layers = model_arch_info.num_layers(model_config)

    kq_spaces = []
    v_spaces = []
    for j in range(num_layers):
        kq_spaces.append(
            _template_substitution(kq_space, num_layers=num_layers, layer_idx=j)
        )
        v_spaces.append(
            _template_substitution(v_space, num_layers=num_layers, layer_idx=j)
        )

    model1_features = safetensors.torch.load_file(model1_ft, device=device)
    model2_features = safetensors.torch.load_file(model2_ft, device=device)

    model1_features.pop("attention_mask")
    model2_features.pop("attention_mask")

    for feature_space in model1_features.keys():
        concatenated_feature = torch.cat(
            (model1_features[feature_space], model2_features[feature_space]), dim=-1
        )

        correlation_matrix = calc_correlation_matrix(concatenated_feature)

        if feature_space in (kq_spaces + v_spaces):
            merge, unmerge = match_tensors_permute_MHA(
                correlation_matrix=correlation_matrix,
                n_heads=model_config.num_attention_heads,
                absval=absval,
            )

        else:
            merge, unmerge = match_tensors_permute(
                correlation_matrix=correlation_matrix,
                absval=absval,
            )

        safetensors.torch.save_file(
            {feature_space: merge.contiguous()},
            f"{out_path}/{feature_space}_merge.safetensor",
        )

        safetensors.torch.save_file(
            {feature_space: unmerge.contiguous()},
            f"{out_path}/{feature_space}_unmerge.safetensor",
        )

        del merge, unmerge, correlation_matrix, concatenated_feature


if __name__ == "__main__":
    main()



import os
import tempfile
import threading

import pytest
import torch

from mergekitty.io import TensorWriter


class TestTensorWriter:
    def test_safetensors(self):
        with tempfile.TemporaryDirectory() as d:
            writer = TensorWriter(d, safe_serialization=True)
            writer.save_tensor("steve", torch.randn(4))
            writer.finalize()

            assert os.path.exists(os.path.join(d, "model-00001-of-00001.safetensors"))
            assert os.path.exists(os.path.join(d, "model.safetensors.index.json"))

    def test_pickle(self):
        with tempfile.TemporaryDirectory() as d:
            writer = TensorWriter(d, safe_serialization=False)
            writer.save_tensor("timothan", torch.randn(4))
            writer.finalize()

            assert os.path.exists(os.path.join(d, "pytorch_model-00001-of-00001.bin"))
            assert os.path.exists(os.path.join(d, "pytorch_model.bin.index.json"))

    def test_duplicate_tensor(self):
        with tempfile.TemporaryDirectory() as d:
            writer = TensorWriter(d, safe_serialization=True)
            jim = torch.randn(4)
            writer.save_tensor("jim", jim)
            writer.save_tensor("jimbo", jim)
            writer.finalize()

            assert os.path.exists(os.path.join(d, "model-00001-of-00001.safetensors"))
            assert os.path.exists(os.path.join(d, "model.safetensors.index.json"))

    def test_flush_is_async(self):
        with tempfile.TemporaryDirectory() as d:
            writer = TensorWriter(d, max_shard_size=40, safe_serialization=True)
            started = threading.Event()
            release = threading.Event()
            errors = []

            def fake_write_shard_file(shard_name, shard):
                started.set()
                assert release.wait(timeout=5)
                with open(os.path.join(d, shard_name), "wb"):
                    pass

            writer._write_shard_file = fake_write_shard_file

            tensor = torch.randn(8)
            writer.save_tensor("a", tensor)

            finished = threading.Event()

            def _save_second_tensor():
                try:
                    writer.save_tensor("b", tensor)
                except Exception as exc:  # pragma: no cover - asserted below
                    errors.append(exc)
                finally:
                    finished.set()

            thread = threading.Thread(target=_save_second_tensor, daemon=True)
            thread.start()

            assert started.wait(timeout=5), "Shard write did not start"
            assert finished.wait(timeout=1), "save_tensor blocked on shard write"
            assert not errors

            release.set()
            thread.join(timeout=5)
            assert not thread.is_alive()
            writer.finalize()

    def test_background_write_failure_surfaces(self):
        with tempfile.TemporaryDirectory() as d:
            writer = TensorWriter(d, max_shard_size=40, safe_serialization=True)

            def fake_write_shard_file(shard_name, shard):
                raise OSError("disk full")

            writer._write_shard_file = fake_write_shard_file

            tensor = torch.randn(8)
            writer.save_tensor("a", tensor)
            writer.save_tensor("b", tensor)

            with pytest.raises(RuntimeError, match="Background shard write failed"):
                writer.finalize()



# Partly Copyright (C) 2025 Arcee AI
# Partly Copyright (C) 2025-2026 Allura-org
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import json
import logging
import os
from typing import Any, Dict, List, Optional, Tuple

import bitsandbytes as bnb
import click
import torch
from peft.tuners.lora import QuantLinear
from safetensors.torch import save_file
from tqdm import tqdm
from transformers import AutoModelForCausalLM
from transformers.pytorch_utils import Conv1D

from mergekitty.architecture import WeightInfo, get_architecture_info
from mergekitty.card import generate_card_lora
from mergekitty.common import ModelReference
from mergekitty.io import LazyTensorLoader


def low_rank_decomposition(
    weight: torch.Tensor, max_rank: int
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Decompose a 2D matrix into low-rank matrices L and R using SVD.

    :param weight: The matrix to decompose, of shape (H, W)
    :param max_rank: The maximum rank of the decomposition
    :return: A tuple of tensors (L, R)
    """
    assert weight.dim() == 2, (
        f"Only support 2D matrix, but input has {weight.dim()} dimensions."
    )
    assert max_rank >= 1, (
        f"Maximum rank must be a positive integer, but input max_rank={max_rank}."
    )

    dtype = weight.dtype

    U, S, Vh = torch.linalg.svd(weight.float(), full_matrices=False)

    final_rank = min(min(weight.shape), max_rank)

    # Distribute S to both to improve numerical precision.
    sqrt_S = torch.sqrt(torch.diag(S[:final_rank]))
    L = sqrt_S @ Vh[:final_rank, :]
    R = U[:, :final_rank] @ sqrt_S

    return L.to(dtype), R.to(dtype)


def decompose_delta_weight(
    base_weight: torch.Tensor,
    finetuned_weight: torch.Tensor,
    max_rank: int,
    device: Optional[str] = None,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Decompose the delta weight into low-rank matrices L and R.

    :param new_weight: The updated weight matrix after applying LoRA
    :param base_weight: The original weight matrix before LoRA
    :param max_rank: The maximum rank for the low-rank decomposition
    :param device: The device to perform computation on
    :return: A tuple of tensors (L, R)
    """
    assert base_weight.size() == finetuned_weight.size(), (
        f"Mismatched dimensions: {base_weight.size()} != {finetuned_weight.size()}"
    )

    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"

    base_weight = base_weight.to(device)
    finetuned_weight = finetuned_weight.to(device)

    delta_weight = finetuned_weight - base_weight

    L, R = low_rank_decomposition(delta_weight, max_rank)

    return L, R


def get_model_details(
    model_id: str,
    skip_undecomposable: bool,
    modules_to_save: Optional[List[str]] = None,
) -> List[Tuple[str, str, torch.Size]]:
    """
    Retrieve architectural details of a given pre-trained model.

    :param model_id: The identifier of the pre-trained model to load
    :param skip_undecomposable: Skip saving undecomposable modules
    :return: A list of tuples where each tuple contains:
             - type: The type of the module ('embedding', 'linear', or 'to_save')
             - name: The full name of the module
             - size: The dimensions of the module's weight tensor
    """

    # Avoid loading weights as we won't need them
    pretrained_model = AutoModelForCausalLM.from_pretrained(
        model_id, state_dict={}, device_map="meta"
    )

    vocab_size = pretrained_model.get_input_embeddings().weight.shape[0]

    module_details = []
    modules_to_save = set(modules_to_save or [])

    for name, module in pretrained_model.named_modules():
        if name in modules_to_save or (
            "." in name and name.split(".")[-1] in modules_to_save
        ):
            module_details.append(("to_save", name, module.weight.size()))
        elif module == pretrained_model.get_input_embeddings():
            # if isinstance(module, torch.nn.Embedding):
            module_details.append(("embedding", name, module.weight.size()))
        elif module == pretrained_model.get_output_embeddings():
            # if isinstance(module, torch.nn.Embedding):
            module_details.append(("output", name, module.weight.size()))
        elif hasattr(module, "weight") and isinstance(module.weight, torch.Tensor):
            if (
                # SEE: https://github.com/huggingface/peft/blob/main/src/peft/tuners/lora/model.py
                isinstance(
                    module,
                    (
                        torch.nn.Linear,
                        torch.nn.Conv2d,
                        bnb.nn.Linear4bit,
                        bnb.nn.Linear8bitLt,
                        QuantLinear,
                        Conv1D,
                    ),
                )
                or (
                    "Linear" in module.__class__.__name__
                    and module.__class__.__name__
                    not in ("LlamaLinearScalingRotaryEmbedding",)
                )
            ):
                module_details.append(("linear", name, module.weight.size()))
            elif not skip_undecomposable:
                module_details.append(("to_save", name, module.weight.size()))
            else:
                logging.info(f"Skipping undecomposable module '{name}'.")

    return module_details, vocab_size


def validate_and_combine_details(
    base_model_id: str,
    finetuned_model_id: str,
    skip_undecomposable: bool,
    extend_vocab: bool,
    modules_to_save: Optional[List[str]] = None,
) -> Tuple[List[Tuple[str, str]], int, int]:
    """
    Validate and combine details from a base model and a fine-tuned model.

    :param base_model_id: The identifier for the base model
    :param finetuned_model_id: The identifier for the fine-tuned model
    :param skip_undecomposable: Skip saving undecomposable modules
    :return: A list of tuples with the type and name of the validated/combined model layers
    """

    base_model_details, base_vocab_size = get_model_details(
        base_model_id, skip_undecomposable, modules_to_save=modules_to_save
    )
    finetuned_model_details, finetuned_vocab_size = get_model_details(
        finetuned_model_id, skip_undecomposable, modules_to_save=modules_to_save
    )

    module_details = []

    for i, (base_layer, finetuned_layer) in enumerate(
        zip(base_model_details, finetuned_model_details)
    ):
        base_type, base_name, base_size = base_layer
        finetuned_type, finetuned_name, finetuned_size = finetuned_layer

        assert base_type == finetuned_type, (
            f"Layer type mismatch: {base_type} != {finetuned_type}"
        )
        assert base_name == finetuned_name, (
            f"Layer name mismatch: {base_name} != {finetuned_name}"
        )

        # Fine-tuned models with added vocab will have have their extra rows truncated unless `extend_vocab` is specified
        if base_type != "to_save" and finetuned_size[0] > base_size[0]:
            assert base_size[1] == finetuned_size[1], (
                f"Column dimension mismatch in layer '{base_name}': {base_size} != {finetuned_size}"
            )

            if base_type == "embedding" or base_type == "output":
                if not extend_vocab:
                    logging.warning(
                        f"Finetuned module '{base_name}' will have {finetuned_size[0] - base_size[0]} rows truncated for weight decomposition! To preserve all embeddings, invoke script with --extend-vocab or --save-module={base_name}."
                    )
                else:
                    logging.warning(
                        f"Base module '{base_name}' will have {finetuned_size[0] - base_size[0]} rows added for weight decomposition. Make sure to call `model.resize_token_embeddings({finetuned_size[0]})` before applying LoRA for inference!"
                    )
            else:
                logging.warning(
                    f"Finetuned module '{base_name}' will have {finetuned_size[0] - base_size[0]} rows truncated for weight decomposition!"
                )

        elif base_type != "to_save":
            assert base_size == finetuned_size, (
                f"Dimension mismatch in layer '{base_name}': {base_size} != {finetuned_size}"
            )

        module_details.append((base_type, base_name))

    return module_details, base_vocab_size, finetuned_vocab_size


def build_wi_map(base_model_ref: ModelReference, trust_remote_code: bool = False):
    weight_info_map = {}
    base_cfg = base_model_ref.config(trust_remote_code=trust_remote_code)
    try:
        arch_info = get_architecture_info(base_cfg)
    except RuntimeError as e:
        logging.error(
            f"Failed to load architecture info for model {base_model_ref}: {e}"
        )
        return {}
    for weight_info in arch_info.all_weights(base_cfg):
        weight_info_map[weight_info.name] = weight_info
    return weight_info_map


def load_weights(
    wi_map: Dict[str, WeightInfo],
    base_loader: LazyTensorLoader,
    finetuned_loader: LazyTensorLoader,
    module_name: str,
):
    optional = False
    aliases = None
    tied_names = None
    if weight_info := wi_map.get(module_name + ".weight"):
        if weight_info.optional:
            optional = True
        if weight_info.aliases:
            aliases = weight_info.aliases
        if weight_info.tied_names:
            tied_names = weight_info.tied_names

    base_weight = base_loader.get_tensor(
        f"{module_name}.weight", aliases=aliases, raise_on_missing=False
    )
    finetuned_weight = finetuned_loader.get_tensor(
        f"{module_name}.weight", aliases=aliases, raise_on_missing=False
    )
    if optional and (base_weight is None and finetuned_weight is None):
        return None, None
    if tied_names:
        if base_weight is None:
            base_weight = base_loader.get_tensor(
                f"{module_name}.weight", aliases=tied_names, raise_on_missing=False
            )
        if finetuned_weight is None:
            finetuned_weight = finetuned_loader.get_tensor(
                f"{module_name}.weight", aliases=tied_names, raise_on_missing=False
            )
    if base_weight is None:
        raise RuntimeError(f"Missing base weight for {module_name}")
    if finetuned_weight is None:
        if optional:
            return None, None
        raise RuntimeError(f"Missing finetuned weight for {module_name}")
    return base_weight, finetuned_weight


def extract_lora(
    module_details: List[Tuple[str, str]],
    base_model_ref: ModelReference,
    finetuned_model_ref: ModelReference,
    max_rank: int,
    extend_vocab: bool,
    no_lazy_unpickle: bool,
    device: Optional[str],
    trust_remote_code: bool = False,
) -> Tuple[Dict[str, torch.Tensor], Dict[str, int]]:
    """
    Process module details to decompose weights and generate LoRA weights and ranks.

    :param module_details: List of module details.
    :param base_model_ref: Reference to the base model.
    :param finetuned_model_ref: Reference to the fine-tuned model.
    :param max_rank: The maximum rank for the low-rank decomposition.
    :param no_lazy_unpickle: Flag to disable lazy unpickle.
    :param device: The device to perform computation on.
    :return: A tuple containing LoRA weights dictionary and ranks dictionary.
    """

    base_loader = LazyTensorLoader(
        base_model_ref.tensor_index(), lazy_unpickle=(not no_lazy_unpickle)
    )
    finetuned_loader = LazyTensorLoader(
        finetuned_model_ref.tensor_index(), lazy_unpickle=(not no_lazy_unpickle)
    )

    lora_weights = {}
    ranks = {}

    wi_map = build_wi_map(base_model_ref, trust_remote_code)

    for module_type, module_name in tqdm(module_details):
        base_weight, finetuned_weight = load_weights(
            wi_map, base_loader, finetuned_loader, module_name
        )
        if base_weight is None and finetuned_weight is None:
            logging.info(f"[{module_type}] {module_name}: optional weight not found")
            continue

        if module_type == "to_save":
            lora_weights[f"base_model.model.{module_name}.weight"] = (
                finetuned_weight.to("cpu").contiguous()
            )

            logging.info(
                f"[{module_type}] {module_name}: output_dims=({finetuned_weight.shape})"
            )

        else:
            if finetuned_weight.shape[0] > base_weight.shape[0]:
                if extend_vocab:
                    print(f"Extra tokens found!, module name : {module_name}")

                    new_base_weight = torch.empty(
                        finetuned_weight.shape, device=base_weight.device
                    )
                    new_base_weight.normal_(mean=0.0, std=0.02)

                    # Copy original base_weight values into the new tensor
                    new_base_weight[: base_weight.shape[0]] = base_weight

                    if module_type == "embedding" or module_type == "output":
                        lora_weights[
                            f"base_model.model.{module_name}.base_layer.weight"
                        ] = new_base_weight.to("cpu").contiguous()

                    base_weight = new_base_weight
                else:
                    logging.warning(
                        f"Finetuned module '{module_name}' will have {finetuned_weight.shape[0] - base_weight.shape[0]} rows truncated for weight decomposition!"
                    )
                    finetuned_weight = finetuned_weight[: base_weight.shape[0]]

            if module_type == "embedding":
                # These need to be transposed for some reason...
                lora_embedding_A, lora_embedding_B = decompose_delta_weight(
                    base_weight.T, finetuned_weight.T, max_rank, device=device
                )

                lora_weights[f"base_model.model.{module_name}.lora_embedding_A"] = (
                    lora_embedding_A.to("cpu").contiguous()
                )
                lora_weights[f"base_model.model.{module_name}.lora_embedding_B"] = (
                    lora_embedding_B.to("cpu").contiguous()
                )

                ranks[module_name] = lora_embedding_A.shape[0]

                logging.info(
                    f"[{module_type}] {module_name}: final_rank={ranks[module_name]}, "
                    f"input_dims=({base_weight.shape}), "
                    f"output_dims=({lora_embedding_A.shape}, {lora_embedding_B.shape})"
                )

            else:
                lora_A, lora_B = decompose_delta_weight(
                    base_weight, finetuned_weight, max_rank, device=device
                )

                lora_weights[f"base_model.model.{module_name}.lora_A.weight"] = (
                    lora_A.to("cpu").contiguous()
                )
                lora_weights[f"base_model.model.{module_name}.lora_B.weight"] = (
                    lora_B.to("cpu").contiguous()
                )

                ranks[module_name] = lora_A.shape[0]

                logging.info(
                    f"[{module_type}] {module_name}: final_rank={ranks[module_name]}, "
                    f"input_dims=({base_weight.shape}), "
                    f"output_dims=({lora_A.shape}, {lora_B.shape})"
                )

    return lora_weights, ranks


def reconstruct_invocation(args: Dict[str, Any]) -> str:
    """
    Reconstruct the command-line invocation string based on the given arguments.

    :param args: A dictionary containing the command arguments with keys matching the parameter names.
                 Expected keys are 'base_model', 'finetuned_model', 'out_path', 'no_lazy_unpickle',
                 'skip_undecomposable, 'max_rank', 'model_name', 'device' and 'verbose'.
    :return: The reconstructed command-line invocation string.
    """

    # Provide a default value for out_path if it's not in the dictionary
    out_path = args.get("out_path", "OUTPUT_PATH")

    invocation = f"mergekitty-extract-lora {args['finetuned_model']} {args['base_model']} {out_path}"
    if args.get("no_lazy_unpickle"):
        invocation += " --no-lazy-unpickle"
    if args.get("skip_undecomposable"):
        invocation += " --skip-undecomposable"
    if args.get("max_rank"):
        invocation += f" --rank={args['max_rank']}"
    if args.get("extend_vocab"):
        invocation += " --extend-vocab"
    if args.get("model_name"):
        invocation += f" --model_name={args['model_name']}"
    if args.get("device"):
        invocation += f" --device={args['device']}"
    if args.get("verbose"):
        invocation += " --verbose"
    if args.get("modules_to_save"):
        for module in args["modules_to_save"]:
            invocation += f" --save-module={module}"

    return invocation


def create_peft_config(
    base_model_name_or_path: str,
    rank: int,
    alpha: int,
    rank_pattern: Dict[str, int],
    alpha_pattern: Dict[str, int],
    target_modules: List[str],
    modules_to_save: List[str],
) -> Dict[str, Any]:
    """
    Create a PEFT (Parameter-Efficient Fine-Tuning) configuration dictionary.

    :param base_model_name_or_path: The path or name of the base model.
    :param rank: The rank for the low-rank adaptation.
    :param alpha: The scaling factor for low-rank adaptation.
    :param rank_pattern: A dictionary specifying rank patterns for different modules.
    :param alpha_pattern: A dictionary specifying alpha patterns for different modules.
    :param target_modules: A list of module names to apply the adaptation to.
    :param modules_to_save: A list of module names to save during the adaptation.
    :return: A dictionary containing the PEFT configuration.
    """
    return {
        "alpha_pattern": alpha_pattern,
        "auto_mapping": None,
        "base_model_name_or_path": base_model_name_or_path,
        "bias": "none",
        "fan_in_fan_out": False,
        "inference_mode": True,
        "init_lora_weights": True,
        "layers_pattern": None,
        "layers_to_transform": None,
        "loftq_config": {},
        "lora_alpha": alpha,
        "lora_dropout": 0,
        "megatron_config": None,
        "megatron_core": "megatron.core",
        "modules_to_save": modules_to_save,
        "peft_type": "LORA",
        "r": rank,
        "rank_pattern": rank_pattern,
        "revision": None,
        "target_modules": target_modules,
        "task_type": "CAUSAL_LM",
        "use_rslora": False,
    }


def save_model_and_config(
    lora_weights: Dict[str, torch.Tensor],
    ranks: Dict[str, int],
    extended: bool,
    embedding_size: int,
    module_details: List[Tuple[str, str]],
    invocation_args: Dict[str, Any],
) -> None:
    """
    Save the PEFT model and configuration to the specified output path.

    :param lora_weights: The LoRA weights.
    :param ranks: The ranks of the LoRA weights.
    :param module_details: Details of the model modules.
    :param invocation_args: The command-line invocation arguments.
    """

    base_model_ref = ModelReference.parse(invocation_args["base_model"])
    finetuned_model_ref = ModelReference.parse(invocation_args["finetuned_model"])
    out_path = invocation_args["out_path"]
    model_name = invocation_args["model_name"]

    # Work out the actual final rank and only retain those that were lower.
    final_max_rank = max(ranks.values())
    ranks = {k: v for k, v in ranks.items() if v != final_max_rank}

    lora_config = create_peft_config(
        base_model_name_or_path=base_model_ref.model.path,
        rank=final_max_rank,
        alpha=final_max_rank,  # Setting the alpha to the rank value as `peft` will scale the LoRA weights by alpha/r when applying the adapter
        rank_pattern=ranks,
        alpha_pattern=ranks,
        target_modules=list(
            set(
                module_name.split(".")[-1]
                for module_type, module_name in module_details
                if module_type != "to_save"
            )
        ),
        modules_to_save=list(
            set(
                module_name.split(".")[-1]
                for module_type, module_name in module_details
                if module_type == "to_save"
            )
        ),
    )

    with open(os.path.join(out_path, "adapter_config.json"), "w") as f:
        json.dump(lora_config, f, indent=2)

    save_file(lora_weights, os.path.join(out_path, "adapter_model.safetensors"))

    invocation_args.pop("out_path")  # don't include out_path for privacy
    invocation = reconstruct_invocation(invocation_args)

    card_md = generate_card_lora(
        base_model_ref=base_model_ref,
        finetuned_model_ref=finetuned_model_ref,
        invocation=invocation,
        extended=extended,
        vocab_size=embedding_size,
        name=model_name,
    )

    with open(os.path.join(out_path, "README.md"), "w", encoding="utf-8") as fp:
        fp.write(card_md)

    logging.info(f"PEFT LoRA adapters saved to {out_path}")


@click.command("mergekitty-extract-lora")
@click.argument("finetuned_model", type=str)
@click.argument("base_model", type=str)
@click.argument("out_path", type=click.Path())
@click.option(
    "--no-lazy-unpickle",
    type=bool,
    is_flag=True,
    default=False,
    help="Disable lazy unpickler (more stable, higher memory usage)",
)
@click.option(
    "--skip-undecomposable",
    type=bool,
    is_flag=True,
    default=False,
    help="Skip saving undecomposable modules in the LoRA",
)
@click.option(
    "--rank",
    "max_rank",
    type=int,
    default=32,
    help="The maximum rank for the low-rank decomposition",
)
@click.option(
    "--extend-vocab",
    is_flag=True,
    default=False,
    help="Extend vocabulary for models with additional tokens instead of truncating",
)
@click.option(
    "--model_name",
    type=str,
    default=None,
    help="Name of the resulting model (shown in the model card)",
)
@click.option(
    "--device",
    type=str,
    default=None,
    help="PyTorch device to perform SVD computation on",
)
@click.option(
    "--verbose", "-v", type=bool, is_flag=True, default=False, help="Verbose logging"
)
@click.option(
    "--save-module",
    "modules_to_save",
    type=str,
    multiple=True,
    default=[],
    help="Save the specified module(s) at full rank",
)
@click.option(
    "--trust-remote-code/--no-trust-remote-code",
    default=False,
    help="Trust remote code when loading model configurations",
)
def main(
    finetuned_model: str,
    base_model: str,
    out_path: str,
    no_lazy_unpickle: bool,
    skip_undecomposable: bool,
    max_rank: int,
    extend_vocab: bool,
    model_name: str,
    device: str,
    verbose: bool,
    modules_to_save: List[str],
    trust_remote_code: bool,
) -> None:
    """
    Decomposes delta weights between a base model and a finetuned model, saving a PEFT model to the specified output path.

    \b
    Arguments:
    FINETUNED_MODEL - the model ID or path to use as the PEFT extraction target model.
    BASE_MODEL - the model ID or path to use as the base model.
    OUT_PATH - the output path where the PEFT model will be saved.
    """

    invocation_args = {
        "base_model": base_model,
        "finetuned_model": finetuned_model,
        "max_rank": max_rank,
        "extend_vocab": extend_vocab,
        "device": device,
        "out_path": out_path,
        "model_name": model_name,
        "no_lazy_unpickle": no_lazy_unpickle,
        "skip_undecomposable": skip_undecomposable,
        "verbose": verbose,
        "modules_to_save": modules_to_save or None,
    }

    logging.basicConfig(level=logging.INFO if verbose else logging.WARNING)

    os.makedirs(out_path, exist_ok=True)

    base_model_ref = ModelReference.parse(base_model)
    finetuned_model_ref = ModelReference.parse(finetuned_model)

    (
        module_details,
        base_vocab_size,
        finetuned_vocab_size,
    ) = validate_and_combine_details(
        base_model_ref.model.path,
        finetuned_model_ref.model.path,
        skip_undecomposable,
        extend_vocab,
        modules_to_save=modules_to_save,
    )
    logging.info(f"Base model vocab size: {base_vocab_size}")
    logging.info(f"Finetuned model vocab size: {finetuned_vocab_size}")

    lora_weights, ranks = extract_lora(
        module_details,
        base_model_ref,
        finetuned_model_ref,
        max_rank,
        extend_vocab,
        no_lazy_unpickle,
        device,
        trust_remote_code,
    )

    save_model_and_config(
        lora_weights,
        ranks,
        finetuned_vocab_size > base_vocab_size and extend_vocab,
        finetuned_vocab_size if extend_vocab else base_vocab_size,
        module_details,
        invocation_args,
    )


if __name__ == "__main__":
    main()



# Partly Copyright (C) 2025 Arcee AI
# Partly Copyright (C) 2025-2026 Allura-org
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import os
import re
import threading
from typing import Dict, Optional, Tuple

import torch

from mergekitty.architecture import WeightInfo
from mergekitty.common import (
    AdapterReference,
    ImmutableMap,
    ModelReference,
    dtype_from_name,
)
from mergekitty.task import Task
from mergekitty.io.lazy_tensor_loader import LazyTensorLoader
from mergekitty.io.tensor_writer import TensorWriter
from mergekitty.options import MergeOptions


class LoaderCache:
    loaders: Dict[ModelReference | AdapterReference, LazyTensorLoader] = {}
    lora_cache_dir: Optional[str] = None
    hf_cache_dir: Optional[str] = None
    lazy_unpickle: bool = False
    trust_remote_code: bool = False
    _lock = threading.Lock()

    # singleton instance
    _instance: Optional["LoaderCache"] = None

    def __new__(cls) -> "LoaderCache":
        if cls._instance is None:
            cls._instance = super(LoaderCache, cls).__new__(cls)
        return cls._instance

    def get(self, model: ModelReference | AdapterReference) -> LazyTensorLoader:
        with self._lock:
            if model not in self.loaders:
                if isinstance(model, ModelReference):
                    merged = model.merged(
                        cache_dir=self.lora_cache_dir,
                        trust_remote_code=self.trust_remote_code,
                    )
                    loader = merged.lazy_loader(
                        cache_dir=self.hf_cache_dir, lazy_unpickle=self.lazy_unpickle
                    )
                else:
                    loader = model.lazy_loader(
                        cache_dir=self.hf_cache_dir, lazy_unpickle=self.lazy_unpickle
                    )
                self.loaders[model] = loader
            return self.loaders[model]

    def flush_all(self):
        with self._lock:
            loaders = list(self.loaders.values())
        for loader in loaders:
            loader.flush()

    def setup(self, options: MergeOptions):
        with self._lock:
            self.lora_cache_dir = options.lora_merge_cache
            self.hf_cache_dir = options.transformers_cache
            self.lazy_unpickle = options.lazy_unpickle
            self.trust_remote_code = options.trust_remote_code


shard_name_re = re.compile(r"model\-([0-9]+)-of-([0-9]+)")


def _normalized_shard_name(path: str) -> str:
    name, _ext = os.path.splitext(os.path.basename(path))
    name = name.lower().replace("pytorch_model", "model")
    if m := shard_name_re.search(name):
        frac = int(m.group(1)) / int(m.group(2))
        name = f"model-{int(frac * 100):03d}pct"
    return name


class LoadTensor(Task[Optional[torch.Tensor]]):
    model: ModelReference | AdapterReference
    tensor: str
    dtype: Optional[str] = None
    device: Optional[str] = None
    optional: bool = False
    aliases: Optional[Tuple[str, ...]] = None
    tied_names: Optional[Tuple[str, ...]] = None

    def arguments(self) -> Dict[str, Task]:
        return {}

    def _resolve_name(self, loader: LazyTensorLoader) -> Optional[str]:
        all_names = (
            [self.tensor] + list(self.aliases or []) + list(self.tied_names or [])
        )
        for name in all_names:
            if name in loader.index.tensor_paths:
                return name
        return None

    def execute(self) -> Optional[torch.Tensor]:
        loader = LoaderCache().get(self.model)
        name = self._resolve_name(loader)
        if not name:
            if not self.optional:
                raise RuntimeError(
                    f"Tensor {self.tensor} required but not present in model {self.model}"
                )
            return None

        x = loader.get_tensor(name, device=self.device or "cpu")
        if self.dtype and (dtype := dtype_from_name(self.dtype)) != x.dtype:
            x = x.to(dtype=dtype)
        return x

    def priority(self) -> int:
        return -1000

    def group_label(self) -> Optional[str]:
        loader = LoaderCache().get(self.model)
        name = self._resolve_name(loader)
        if not name:
            return None

        shard_path = loader.index.tensor_paths[name]
        return _normalized_shard_name(shard_path)


class GatherTensors(Task[Dict[ModelReference, torch.Tensor]]):
    weight_info: ImmutableMap[ModelReference, WeightInfo]
    dtype: Optional[str] = None
    device: Optional[str] = None

    def arguments(self) -> Dict[str, Task]:
        return {
            f"{str(model)}:{wi.name}": LoadTensor(
                model=model,
                tensor=wi.name,
                dtype=wi.force_dtype or self.dtype,
                device=self.device,
                optional=wi.optional,
                aliases=wi.aliases,
                tied_names=wi.tied_names,
            )
            for (model, wi) in self.weight_info.items()
        }

    def group_label(self) -> Optional[str]:
        return max(t.group_label() or "" for t in self.arguments().values())

    def priority(self) -> int:
        return -10

    def execute(self, **kwargs) -> Dict[ModelReference, torch.Tensor]:
        key2model = {
            f"{str(model)}:{wi.name}": model for (model, wi) in self.weight_info.items()
        }
        return {
            key2model[key]: kwargs[key] for key in key2model if kwargs[key] is not None
        }


class TensorWriterTask(Task[TensorWriter]):
    out_path: str
    max_shard_size: int
    safe_serialization: bool = True

    def arguments(self) -> Dict[str, Task]:
        return {}

    def execute(self, **_kwargs) -> TensorWriter:
        return TensorWriter(
            self.out_path,
            max_shard_size=self.max_shard_size,
            safe_serialization=self.safe_serialization,
        )


class SaveTensor(Task[None]):
    tensor_name: str
    tensor_task: Task
    writer_task: TensorWriterTask
    clone: bool
    optional: bool = False
    dtype: Optional[str] = None

    def arguments(self) -> Dict[str, Task]:
        return {"writer": self.writer_task, "tensor": self.tensor_task}

    def priority(self) -> int:
        return 1000

    def group_label(self) -> Optional[str]:
        return self.tensor_task.group_label()

    def execute(self, writer: TensorWriter, tensor: Optional[torch.Tensor]) -> None:
        if tensor is None:
            if not self.optional:
                raise RuntimeError(f"No value for required tensor {self.tensor_name}")
            return
        if self.dtype:
            tensor = tensor.to(dtype=dtype_from_name(self.dtype))
        writer.save_tensor(name=self.tensor_name, tensor=tensor, clone=self.clone)


class FinalizeModel(Task[None]):
    tensor_save_tasks: Tuple[Task, ...]
    writer_task: TensorWriterTask

    def arguments(self) -> Dict[str, Task]:
        return {
            "writer": self.writer_task,
            **{f"_unused_{idx}": t for idx, t in enumerate(self.tensor_save_tasks)},
        }

    def execute(self, writer: TensorWriter, **kwargs) -> None:
        writer.finalize()


class BuildStateDict(Task[Dict[str, torch.Tensor]]):
    tensors: ImmutableMap[WeightInfo, Task[torch.Tensor]]

    def arguments(self) -> Dict[str, Task]:
        return {str(wi): t for wi, t in self.tensors.items()}

    def execute(self, **kwargs) -> Dict[str, torch.Tensor]:
        return {str(wi): t for wi, t in self.tensors.items()}


class ReturnTensor(Task[torch.Tensor]):
    weight_info: WeightInfo
    tensor_task: Task[torch.Tensor]
    dtype: Optional[str] = None

    def arguments(self) -> Dict[str, Task]:
        return {"tensor": self.tensor_task}

    def priority(self) -> int:
        return 10000

    def group_label(self) -> Optional[str]:
        return self.tensor_task.group_label()

    def execute(self, tensor: torch.Tensor) -> torch.Tensor:
        if self.dtype and (dtype := dtype_from_name(self.dtype)) != tensor.dtype:
            tensor = tensor.to(dtype=dtype)
        return tensor



# Copyright (C) 2025 Arcee AI
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

from typing import Any, Dict, List, Optional

import torch
from typing_extensions import override

from mergekitty.architecture import WeightInfo
from mergekitty.common import ImmutableMap, ModelReference
from mergekitty.task import Task
from mergekitty.merge_methods.base import (
    ConfigParameterDef,
    MergeMethod,
    MergeTensorInput,
)
from mergekitty.merge_methods.rectify_embed import rectify_embed_sizes


class LinearMergeTask(Task[torch.Tensor]):
    gather_tensors: MergeTensorInput
    tensor_parameters: ImmutableMap[ModelReference, ImmutableMap[str, Any]]
    normalize: bool
    weight_info: WeightInfo

    def uses_accelerator(self) -> bool:
        return True

    def arguments(self) -> Dict[str, Task]:
        return {"tensors": self.gather_tensors}

    def execute(
        self, tensors: Dict[ModelReference, torch.Tensor], **_kwargs
    ) -> torch.Tensor:
        keys = list(tensors.keys())

        tensors = [tensors[key] for key in keys]
        weights = [self.tensor_parameters[key]["weight"] for key in keys]

        rectify_embed_sizes(self.weight_info, tensors)

        unique_shapes = set(t.shape for t in tensors)
        if len(unique_shapes) != 1:
            raise RuntimeError(
                f"Tensor size mismatch for {self.weight_info.name}, sizes: {list(unique_shapes)}"
            )

        tensors = torch.stack(tensors, dim=0)
        weights = torch.tensor(weights, dtype=tensors.dtype, device=tensors.device)
        while len(weights.shape) < len(tensors.shape):
            weights.unsqueeze_(-1)

        res = (weights * tensors).sum(dim=0)
        if self.normalize:
            res = res / weights.sum(dim=0)

        return res

    def group_label(self) -> Optional[str]:
        return self.gather_tensors.group_label()


class LinearMerge(MergeMethod):
    def name(self) -> str:
        return "linear"

    @override
    def pretty_name(self) -> Optional[str]:
        return "Linear"

    @override
    def reference_url(self) -> Optional[str]:
        return "https://arxiv.org/abs/2203.05482"

    def parameters(self) -> List[ConfigParameterDef]:
        return [
            ConfigParameterDef(name="normalize", required=False, default_value=True),
        ]

    def tensor_parameters(self) -> List[ConfigParameterDef]:
        return [ConfigParameterDef(name="weight", required=True)]

    def make_task(
        self,
        *,
        output_weight: WeightInfo,
        tensors: MergeTensorInput,
        parameters: Dict[str, Any],
        tensor_parameters: ImmutableMap[ModelReference, ImmutableMap[str, Any]],
        **_kwargs,
    ) -> Task:
        return LinearMergeTask(
            gather_tensors=tensors,
            tensor_parameters=tensor_parameters,
            normalize=parameters["normalize"],
            weight_info=output_weight,
        )



from typing import List

from mergekitty.moe.arch import MoEOutputArchitecture
from mergekitty.moe.deepseek import DeepseekMoE
from mergekitty.moe.mixtral import MixtralMoE

ALL_OUTPUT_ARCHITECTURES: List[MoEOutputArchitecture] = [MixtralMoE(), DeepseekMoE()]

try:
    from mergekitty.moe.qwen import QwenMoE
except ImportError:
    pass
else:
    ALL_OUTPUT_ARCHITECTURES.append(QwenMoE())

__all__ = [
    "ALL_OUTPUT_ARCHITECTURES",
    "MoEOutputArchitecture",
]



# Partly Copyright (C) 2025 Arcee AI
# Partly Copyright (C) 2025-2026 Allura-org
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import json
import logging
import os
import queue
import threading
from typing import Dict, Optional, Tuple

import safetensors
import torch


class TensorWriter:
    out_path: str
    max_shard_size: int
    shards_written: int
    weight_map = Dict[str, str]
    current_shard: Dict[str, torch.Tensor]
    current_shard_size: int
    total_size: int
    safe_serialization: bool
    _pending_writes: "queue.Queue[Tuple[str, Dict[str, torch.Tensor]] | object]"
    _writer_thread: threading.Thread
    _writer_error: Optional[BaseException]
    _STOP_SENTINEL = object()

    def __init__(
        self,
        out_path: str,
        max_shard_size: int = 1000 * 1000 * 1000 * 5,
        safe_serialization: bool = True,
    ) -> None:
        os.makedirs(out_path, exist_ok=True)

        self.out_path = out_path
        self.max_shard_size = max_shard_size
        self.safe_serialization = safe_serialization
        self.shards_written = 0
        self.weight_map = {}
        self.current_shard = {}
        self.current_shard_size = 0
        self.total_size = 0
        self._lock = threading.RLock()
        self._pending_writes = queue.Queue(maxsize=1)
        self._writer_error = None
        self._writer_thread = threading.Thread(
            target=self._writer_loop,
            name="mergekitty-writer",
            daemon=True,
        )
        self._writer_thread.start()

    def save_tensor(self, name: str, tensor: torch.Tensor, clone: bool = False):
        shard_to_write = None
        with self._lock:
            self._raise_writer_error_locked()
            if not tensor.is_contiguous():
                tensor = tensor.contiguous()

            tensor_size = tensor.numel() * tensor.element_size()
            if (
                self.current_shard
                and self.current_shard_size + tensor_size > self.max_shard_size
            ):
                shard_to_write = self._rotate_shard_locked()

            if clone:
                tensor = tensor.clone()

            self.current_shard[name] = tensor
            self.total_size += tensor_size
            self.current_shard_size += tensor_size
        if shard_to_write is not None:
            self._enqueue_write(shard_to_write)

    def flush_current_shard(self):
        shard_to_write = None
        with self._lock:
            self._raise_writer_error_locked()
            shard_to_write = self._rotate_shard_locked()
        if shard_to_write is not None:
            self._enqueue_write(shard_to_write)

    def finalize(self):
        self.flush_current_shard()
        self._enqueue_write(self._STOP_SENTINEL)
        self._writer_thread.join()

        with self._lock:
            self._raise_writer_error_locked()

            logging.info("Finalizing shard names")

            prefix, extension = self._get_name_components()

            # standardize shard names to hf format
            total_shards = self.shards_written
            name_remap = {}
            for idx in range(total_shards):
                name_remap[f"{prefix}-{idx + 1}.{extension}"] = (
                    f"{prefix}-{idx + 1:05d}-of-{total_shards:05d}.{extension}"
                )

            for old_name, new_name in name_remap.items():
                os.rename(
                    os.path.join(self.out_path, old_name),
                    os.path.join(self.out_path, new_name),
                )

            for key in self.weight_map:
                self.weight_map[key] = name_remap[self.weight_map[key]]

            with open(
                os.path.join(self.out_path, f"{prefix}.{extension}.index.json"),
                "w",
                encoding="utf-8",
            ) as file:
                json.dump(
                    {
                        "metadata": {
                            "mergekitty_version": "0.0.7",
                            "total_size": self.total_size,
                        },
                        "weight_map": self.weight_map,
                    },
                    file,
                )

    def _get_name_components(self):
        if self.safe_serialization:
            return "model", "safetensors"
        return "pytorch_model", "bin"

    def _writer_loop(self):
        while True:
            item = self._pending_writes.get()
            try:
                if item is self._STOP_SENTINEL:
                    return

                shard_name, shard = item
                logging.info("Writing shard %s to disk", shard_name)
                self._write_shard_file(shard_name, shard)
            except BaseException as exc:  # pragma: no cover - surfaced via public API
                with self._lock:
                    if self._writer_error is None:
                        self._writer_error = exc
                return
            finally:
                self._pending_writes.task_done()

    def _enqueue_write(self, item: Tuple[str, Dict[str, torch.Tensor]] | object):
        while True:
            with self._lock:
                self._raise_writer_error_locked()
            try:
                self._pending_writes.put(item, timeout=0.1)
                return
            except queue.Full:
                continue

    def _raise_writer_error_locked(self):
        if self._writer_error is not None:
            raise RuntimeError("Background shard write failed") from self._writer_error

    def _rotate_shard_locked(self) -> Optional[Tuple[str, Dict[str, torch.Tensor]]]:
        if not self.current_shard:
            return None

        prefix, extension = self._get_name_components()
        shard_name = f"{prefix}-{self.shards_written + 1}.{extension}"
        shard = self.current_shard

        for key in shard:
            self.weight_map[key] = shard_name

        self.current_shard = {}
        self.current_shard_size = 0
        self.shards_written = self.shards_written + 1
        return shard_name, shard

    def _write_shard_file(self, shard_name: str, shard: Dict[str, torch.Tensor]):
        shard_path = os.path.join(self.out_path, shard_name)
        if self.safe_serialization:
            self._save_st(shard_path, shard)
        else:
            torch.save(shard, shard_path)

    def _save_st(self, shard_path: str, shard: Dict[str, torch.Tensor]):
        def _do_save():
            safetensors.torch.save_file(
                shard,
                shard_path,
                metadata={"format": "pt"},
            )

        try:
            _do_save()
        except RuntimeError as e:
            if (
                len(e.args) > 0
                and isinstance(e.args[0], str)
                and "share memory" in e.args[0]
            ):
                logging.warning(
                    "Your model has duplicated tensors but the --clone-tensors "
                    "flag is not set."
                )
                shard = {key: shard[key].clone() for key in shard}
                _do_save()
            else:
                raise
